from typing import Optional
from pyexlatex.models.item import SimpleItem


class Inst(SimpleItem):
    name = 'inst'

    def __init__(self, num: int):
        self.num = num
        super().__init__(self.name, str(num))


from pyexlatex.logic.output.api.builders.base import BaseBuilder


class LuaLatexBuilder(BaseBuilder):
    """A simple lualatex-based builder for LaTeX files.

    Builds LaTeX files by copying them to a temporary directly and running
    ``lualatex`` until the associated ``.aux`` file stops changing.

    :param executable: The path to the ``lualatex`` binary (will looked up on
                    ``$PATH``).
    :param max_runs: An integer providing an upper limit on the amount of times
                     ``lualatex`` can be rerun before an exception is thrown.
    """
    output_extension = 'pdf'
    default_executable = 'lualatex'




from typing import Iterable, Sequence

from mixins.repr import ReprMixin
from pyexlatex.table.models.data.dataitem import DataItem
from pyexlatex.table.models.labels.label import Label
from pyexlatex.table.models.mixins.addvalues.row import RowAddMixin
from pyexlatex.table.models.spacing.cell import CellSpacer


class RowBase(ReprMixin, RowAddMixin):
    values: Sequence

    def __len__(self):
        return sum(_get_length(value) for value in self.values)

    def __iter__(self):
        for value in self.values:
            yield value

    def __getitem__(self, item):
        return self.values[item]

    def __eq__(self, other):
        if not hasattr(other, 'values'):
            return False
        if len(self.values) != len(other.values):
            return False

        for val, other_value in zip(self.values, other.values):
            if val != other_value:
                return False

        return True

    def pad(self, length: int, direction='right'):
        """
        Expand row out to the right or left with blanks, until it is length passed
        :param length:
        :return:
        """
        _pad_obj(self, length=length, direction=direction)

    @property
    def is_spacer(self):
        booleans = []
        for item in self.values:
            if isinstance(item, CellSpacer):
                result = True
            elif isinstance(item, (Label, DataItem)):
                result = str(item.value).strip() == ''
            elif isinstance(item, str):
                result = item.strip() == ''
            else:
                raise ValueError(f'cannot check whether {item} of type {type(item)} is a spacer or not')
            booleans.append(result)
        return all(booleans)


def _get_length(obj):
    from pyexlatex.table.models.texgen.lines import TableLineOfSegments, TableLineSegment
    from pyexlatex.table.models.labels.collection import LabelCollection
    if isinstance(obj, (Label, LabelCollection, DataItem, TableLineOfSegments, TableLineSegment)):
        return len(obj)
    else:
        return 1

def _pad_obj(obj, length: int, direction='right'):
    """
    Expand row out to the right or left with blanks, until it is length passed
    :param length:
    :return:
    """
    num_values_to_add = length - len(obj)
    direction = direction.lower().strip()

    if num_values_to_add == 0:
        return

    if direction == 'right':
        obj.values += [CellSpacer(num_values_to_add)]
    elif direction == 'left':
        obj.values = [CellSpacer(num_values_to_add)] + obj.values
    else:
        raise ValueError(f'must pass left or right for direction. got {direction}')


from typing import List, Optional, Sequence, Union, Callable
from pyexlatex.models.document import DocumentBase
from pyexlatex.typing import ItemOrListOfItems
from pyexlatex.models.package import Package
from pyexlatex.models.control.documentclass.documentclass import DocumentClass
from pyexlatex.resume.models.name import Name
from pyexlatex.resume.models.contactline import ContactLine


class Resume(DocumentBase):
    """
    The main high-level class for creating resumes
    """
    name = 'document'

    def __init__(self, content: ItemOrListOfItems, packages: List[Package]=None,
                 pre_env_contents: Optional[ItemOrListOfItems] = None,
                 name: Optional[str] = None, contact_lines: Optional[Sequence[Union[str, Sequence[str]]]] = None,
                 font_size: Optional[float] = 11,
                 page_modifier_str: Optional[str]='left=0.75in,top=0.6in,right=0.75in,bottom=0.6in',
                 authors: Optional[Union[str, Sequence[str]]] = None, pre_output_func: Optional[Callable] = None):

        self.init_data()

        self.document_class_obj = DocumentClass(
            document_type='resume',
            font_size=font_size,
        )

        from pyexlatex.models.item import ItemBase
        if pre_env_contents is None:
            pre_env_contents_list = []
        elif isinstance(pre_env_contents, (ItemBase, str)):
            pre_env_contents_list = [pre_env_contents]
        else:
            pre_env_contents_list = pre_env_contents  # type: ignore

        name = self._get_name(name, authors)

        if name is not None:
            pre_env_contents_list.append(
                Name(name)
            )

        if contact_lines is not None:
            pre_env_contents_list.extend(
                [ContactLine(contact_info) for contact_info in contact_lines]
            )

        if page_modifier_str is not None:
            # Set margins, body size, etc. with geometry package
            self.data.packages.append(Package('geometry', modifier_str=page_modifier_str))

        super().__init__(
            content,
            packages=packages,
            pre_env_contents=pre_env_contents_list,
            pre_output_func=pre_output_func,
        )

    def _get_name(self, name: Optional[str] = None, authors: Optional[Union[str, Sequence[str]]] = None):
        if authors is None:
            return name
        elif name is None:
            # Authors is passed, name is not
            if isinstance(authors, (list, tuple)):
                authors = authors[0]
            return authors

        # Both name and author passed

        if name == authors:
            return name

        # Name and author are not the same
        raise ValueError(f'got different name and authors, not clear which to use. name: {name}, authors: {authors}')




from typing import Sequence, Type

from pyexlatex.models.lists.base import ListBase
from pyexlatex.presentation.beamer.frame.frame import Frame
from pyexlatex.presentation.beamer.templates.lists.dim_reveal_items import DimAndRevealListItems
from pyexlatex.models.lists.ordered import OrderedList
from pyexlatex.models.lists.unordered import UnorderedList


class DimRevealMixin:

    def __init__(self, content: Sequence[str], ordered_list: bool = False, **frame_kwargs):
        list_class: Type[ListBase]
        if ordered_list:
            list_class = OrderedList
        else:
            list_class = UnorderedList
        content = list_class([DimAndRevealListItems(content, vertical_fill=True)])
        super().__init__(content, **frame_kwargs)  # type: ignore


class DimRevealListFrame(DimRevealMixin, Frame):
    """
    A Frame where the content is bulleted or numbered dim and reveal items
    """

    def __init__(self, content: Sequence[str], ordered_list: bool = False, **frame_kwargs):
        super().__init__(content, ordered_list=ordered_list, **frame_kwargs)


from pyexlatex import TableOfContents


def test_toc():
    assert str(TableOfContents()) == '\\tableofcontents'
    assert str(TableOfContents(options=('currentsection',))) == '\\tableofcontents[currentsection]'



import pandas as pd

import pyexlatex as pl
from tests.base import EXAMPLE_IMAGE_PATH, GENERATED_FILES_DIR
from tests.utils.pdf import compare_pdfs_in_generated_vs_input_by_name


def test_single_figure_from_graphic():
    graphic = pl.Graphic(str(EXAMPLE_IMAGE_PATH))
    fig = pl.Figure([graphic])
    assert str(fig) == '\\begin{figure}\n\\includegraphics[width=1.0\\textwidth]{Sources/nd-logo.png}\n\\end{figure}'
    fig = pl.Figure([graphic], position_str='[h!]')
    assert str(fig) == '\\begin{figure}\n[h!]\n\\includegraphics[width=1.0\\textwidth]{Sources/nd-logo.png}\n\\end{figure}'
    fig = pl.Figure([graphic], caption='image')
    assert str(fig) == '\\begin{figure}\n\\includegraphics[width=1.0\\textwidth]{Sources/nd-logo.png}\n\\caption{image}\n\\end{figure}'
    fig = pl.Figure.from_dict_of_names_and_filepaths({'image': str(EXAMPLE_IMAGE_PATH)})
    assert str(fig) == '\\begin{figure}\n\\includegraphics[width=0.45\\linewidth]{Sources/nd-logo.png}\n\\caption{image}\n\\end{figure}'
    fig = pl.Figure([graphic], caption='woo', label='yeah', centering=False, landscape=True, position_str=r'[h!]')
    assert str(fig) == '\\begin{lfigure}\n[h!]\n\\includegraphics[width=1.0\\textwidth]{Sources/nd-logo.png}\n\\caption{woo}\n\\label{yeah}\n\\end{lfigure}'
    name = 'figure from single graphic document'
    fig.to_pdf_and_move(outfolder=GENERATED_FILES_DIR, outname=name)
    compare_pdfs_in_generated_vs_input_by_name(name)


def test_subfigure_graphic():
    subf1 = pl.Subfigure(str(EXAMPLE_IMAGE_PATH))
    subf2 = pl.Subfigure(str(EXAMPLE_IMAGE_PATH))
    fig = pl.Figure([subf1, subf2])
    assert str(fig) == '\\begin{figure}\n\\centering\n\\begin{subfigure}\n[t]{0.45\\linewidth}\n\\centering\n\\includegraphics[width=1.0\\textwidth]{Sources/nd-logo.png}\n\\end{subfigure}\n\\begin{subfigure}\n[t]{0.45\\linewidth}\n\\centering\n\\includegraphics[width=1.0\\textwidth]{Sources/nd-logo.png}\n\\end{subfigure}\n\\end{figure}'


def test_figure_from_matplotlib():
    plot = pd.Series([1, 2, 3, 4]).plot().get_figure()
    fig = pl.Figure.from_dict_of_names_and_plt_figures({'plot': plot}, GENERATED_FILES_DIR)
    assert str(fig) == '\\begin{figure}\n\\includegraphics[width=0.45\\linewidth]{Sources/plot.pdf}\n\\caption{plot}\n\\end{figure}'
    name = 'figure from matplotlib'
    fig.to_pdf_and_move(outfolder=GENERATED_FILES_DIR, outname=name)
    compare_pdfs_in_generated_vs_input_by_name(name)
    fig = pl.Figure.from_dict_of_names_and_plt_figures({'plot': plot}, GENERATED_FILES_DIR, position_str='[h!]')
    assert str(fig) == '\\begin{figure}\n[h!]\n\\includegraphics[width=0.45\\linewidth]{Sources/plot.pdf}\n\\caption{plot}\n\\end{figure}'


def test_graphic_from_single_graphic_figure():
    graphic = pl.Graphic(str(EXAMPLE_IMAGE_PATH))
    fig = pl.Figure([graphic])
    graphics = fig.to_graphic_list()
    assert str(graphic) == str(graphics[0])
    assert len(graphics) == 1


def test_graphics_from_multiple_subfigure_figure():
    graphic = pl.Graphic(str(EXAMPLE_IMAGE_PATH))
    subf1 = pl.Subfigure(str(EXAMPLE_IMAGE_PATH))
    subf2 = pl.Subfigure(str(EXAMPLE_IMAGE_PATH))
    fig = pl.Figure([subf1, subf2])
    graphics = fig.to_graphic_list()
    assert len(graphics) == 2
    for g in graphics:
        assert str(graphic) == str(g)


from typing import Sequence

from pyexlatex.models.template import Template
from pyexlatex.models.item import SimpleItem
from pyexlatex.models.sizes.paperwidth import PaperWidth
from pyexlatex.presentation.beamer.colorbox import BeamerColorBox
from pyexlatex.presentation.beamer.theme.settemplate import SetBeamerTemplate


class InsertNavigation(SimpleItem):
    name = 'insertnavigation'

    def __init__(self, width, **kwargs):
        self.width = width
        super().__init__(self.name, width, **kwargs)


class AddNavigationHeader(Template):

    def __init__(self, options: Sequence[str] = ('ht=2.25ex', 'dp=3.75ex')):
        insert_nav = InsertNavigation(PaperWidth())
        color_box = BeamerColorBox(insert_nav, 'section in head/foot', options=options)
        set_template = SetBeamerTemplate('headline', color_box)
        self.contents = set_template
        super().__init__()



from pyexlatex.models.item import SimpleItem
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.typing import PyexlatexItems


class PS(TextAreaMixin, SimpleItem):
    name = 'ps'

    def __init__(self, ps: PyexlatexItems):
        self.ps = ps
        super().__init__(self.name, self.ps)



from pyexlatex.models.page.footer import RightFooter
from pyexlatex.models.page.header import FancyHeader
from pyexlatex.models.page.style import PageStyle, ThisPageStyle, FancyPageStyle
from tests.base import SimpleItemTest, MultiOptionSimpleItemTest


class TestPageStyle(SimpleItemTest):
    item_class = PageStyle
    tag_name = 'pagestyle'


class TestThisPageStyle(SimpleItemTest):
    item_class = ThisPageStyle
    tag_name = 'thispagestyle'


class TestFancyPageStyle:

    def test_str(self):
        content = FancyPageStyle('woo')
        assert str(content) == '\\fancypagestyle{plain}{woo}'

    def test_list(self):
        content = FancyPageStyle(['woo', FancyHeader('stuff'), RightFooter('yeah')])
        assert str(content) == '\\fancypagestyle{plain}{woo\n\\fancyhead{stuff}\n\\rfoot{yeah}}'




from pyexlatex.models.item import SimpleItem
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.typing import PyexlatexItems


class Opening(TextAreaMixin, SimpleItem):
    name = 'opening'

    def __init__(self, opening: PyexlatexItems):
        self.opening = opening
        super().__init__(self.name, self.opening)



from typing import Optional, Sequence
from pandas import DataFrame
import pandas as pd
from pyexlatex.table.models.data.row import DataRow
from pyexlatex.table.models.data.dataitem import DataItem
from pyexlatex.table.models.table.section import TableSection


class ValuesTable(TableSection):
    """
    Python interface for a part of a table containing rows and columns with values
    """

    def __init__(self, rows: Sequence[DataRow], break_size_adjustment: Optional[str] = None):
        super().__init__(rows, break_size_adjustment)

    @classmethod
    def from_df(cls, df):
        latex_str_list = _build_latex_str_list_from_df(df)
        rows = [DataRow.from_latex_row_str(row_str) for row_str in latex_str_list]
        return cls(rows)

    @classmethod
    def from_list_of_lists(cls, list_of_lists):
        data_rows: [DataRow] = []
        for list_row in list_of_lists:
            data_values = [DataItem(value) for value in list_row]
            data_rows.append(
                DataRow(data_values)
            )

        return cls(data_rows)

    def __add__(self, other):
        # import here to avoid circular imports
        from pyexlatex.table.models.spacing.columntable import ColumnPadTable

        table_section: TableSection = super().__add__(other)

        # if just adding padding, retain class rather than drop to general base TableSection
        if isinstance(other, ColumnPadTable):
            return ValuesTable(table_section.rows)
        else:
            return table_section

    def __repr__(self):
        return f'<ValuesTable(shape=({len(self.rows)} , {self.num_columns}))>'


def _build_latex_str_list_from_df(df: DataFrame):
    # TODO [#23]: remvove setting of max col width once pandas does this by default for to_latex
    #
    # See https://github.com/pandas-dev/pandas/issues/6491
    with pd.option_context("max_colwidth", None):
        latex_str = df.to_latex(header=False, index=False)
    latex_list = latex_str.split('\n')

    # for a two row dataframe, latex list will be in the format
    # ['\\begin{tabular}{ll}',
    #  '\\toprule',
    #  ' 03790A & 2011-07-31 \\\\',
    #  ' 03790A & 2011-10-31 \\\\',
    #  '\\bottomrule',
    #  '\\end{tabular}',
    #  '']
    latex_list = latex_list[2:-3] # now contains only data rows

    return latex_list




from pyexlatex.models.page.footer import LeftFooter, RightFooter, CenterFooter, FooterLine
from tests.base import SimpleItemTest


class TestLeftFooter(SimpleItemTest):
    item_class = LeftFooter
    tag_name = 'lfoot'


class TestRightFooter(SimpleItemTest):
    item_class = RightFooter
    tag_name = 'rfoot'


class TestCenterFooter(SimpleItemTest):
    item_class = CenterFooter
    tag_name = 'cfoot'


class TestFooterLine:

    def test_no_args(self):
        contents = FooterLine()
        assert str(contents) == '\\renewcommand{\\footrulewidth}{0.4pt}'

    def test_args(self):
        contents = FooterLine(1, 'cm')
        assert str(contents) == '\\renewcommand{\\footrulewidth}{1cm}'



from typing import Sequence, Any, Union

from pyexlatex.logic.format.sizing import adjust_to_full_size_and_center, adjust_to_size
from pyexlatex.presentation.beamer.frame.frame import Frame
from pyexlatex.figure.models.graphic import Graphic
from pyexlatex.models.format.fills import VFill, HFill
from pyexlatex.models.format.centering import Center


class FullWidthFrame(Frame):
    """
    Resizes passed latex object to take up entire frame.
    """

    def __init__(self, content: Sequence[Sequence[Any]], **kwargs):
        content = adjust_to_full_size_and_center(content)
        super().__init__(content, **kwargs)


class GraphicFrame(FullWidthFrame):
    """
    Resizes passed graphic to take up entire frame. Can pass a file path or a latex object.
    """

    def __init__(self, content: Any, **kwargs):
        if isinstance(content, str):
            content = Graphic(content)
        super().__init__(content, **kwargs)


class MultiGraphicFrame(Frame):
    """
    Resizes each graphic to full width and puts vertical space in between graphics.
    Can pass a file path or a latex object.
    """
    HORIZTONAL_SPACING = 0.05
    VERTICAL_SPACING = 0.05
    MAX_WIDTH = 0.9
    MAX_HEIGHT = 0.8

    def __init__(self, content: Sequence[Sequence[Any]], vertical: bool = True, **kwargs):
        self.vertical = vertical
        self.num_contents = len(content)
        all_content = []
        for cont in content:
            if isinstance(cont, str):
                cont = Graphic(cont)
            cont = adjust_to_size(cont, width=self._graphic_width, height=self._graphic_height)
            if self.vertical:
                cont = Center(cont)
            all_content.extend([cont, self._spacer_obj])
        all_content = all_content[:-1]  # strip final spacer
        super().__init__(all_content, **kwargs)

    @property
    def _graphic_width(self) -> float:
        if self.vertical:
            return self.MAX_WIDTH

        num_spacers = self.num_contents - 1
        spacer_space = self.HORIZTONAL_SPACING * num_spacers
        available_space = self.MAX_WIDTH - spacer_space
        space_per_graphic = available_space / self.num_contents

        return space_per_graphic

    @property
    def _graphic_height(self) -> float:
        if not self.vertical:
            return self.MAX_HEIGHT

        num_spacers = self.num_contents - 1
        spacer_space = self.VERTICAL_SPACING * num_spacers
        available_space = self.MAX_HEIGHT - spacer_space
        space_per_graphic = available_space / self.num_contents

        return space_per_graphic

    @property
    def _spacer_obj(self) -> Union[VFill, HFill]:
        if self.vertical:
            return VFill()

        return HFill()







from typing import Optional, Sequence, Union, List
from pyexlatex.models.item import SimpleItem
from pyexlatex.models.credits.institution.inst import Inst
from pyexlatex.models.credits.institution.institute import Institutes


class Author(SimpleItem):
    name = 'author'
    institution_nums: List[int]

    def __init__(self, authors: Union[str, Sequence[str]], short_author: Optional[str] = 'auto',
                 institutions: Optional[Sequence[Sequence[str]]] = None, short_institution: Optional[str] = None):
        """

        :param authors:
        :param short_author: pass short name for author, 'auto' to automatically use first author as short author,
            or None for there to be no short name
        :param institutions:
        :param short_institution:
        """
        if isinstance(authors, str):
            authors = [authors]
        self.authors = authors
        self.short_author = short_author
        self.institutions = institutions
        self.short_institution = short_institution
        self._validate()
        self.institution_nums = []
        self._setup_institutions()
        self.content = self._get_content()

        super().__init__(self.name, self.content, pre_modifiers=self._short_author_modifiers_str)

    @property
    def _short_author_modifiers_str(self) -> Optional[str]:
        if self.short_author is None:
            return None

        if self.short_author.lower() == 'auto':
            short_author = self.authors[0]
        else:
            short_author = self.short_author

        return self._wrap_with_bracket(short_author)


    def _get_content(self):
        if self.institutions is None:
            return self.authors

        content = []
        for i, author in enumerate(self.authors):
            content.append(author + Inst(self.institution_nums[i]))
        return ', '.join(content)

    def _validate(self):
        if self.institutions is not None:
            self._check_lengths_match_raise_value_error_if_not(
                self.authors,
                self.institutions,
                'authors',
                'instutitions'
            )

    def _setup_institutions(self):
        if self.institutions is None:
            return
        # Number institutions starting with 1, but if institution already has a number then use that again
        count = 0
        counted_institutions = {}
        for institution in self.institutions:
            institution = tuple(institution)
            if institution not in counted_institutions:
                count += 1
                counted_institutions[institution] = count
            self.institution_nums.append(counted_institutions[institution])

        self._check_lengths_match_raise_value_error_if_not(
            self.institutions,
            self.institution_nums,
            'instutitions',
            'institution_nums'
        )

        self.init_data()
        institutes_def = Institutes(self.institutions, short_institution=self.short_institution)
        self.data.packages.append(institutes_def)

    def _check_lengths_match_raise_value_error_if_not(self, item1: Sequence, item2: Sequence,
                                                      item1_name: str, item2_name: str):
        len_item_1 = len(item1)
        len_item_2 = len(item2)
        if len_item_1 != len_item_2:
            raise ValueError(f'number of {item1_name} must match that of {item2_name}. got {len_item_1} '
                             f'{item1_name} and {len_item_2} {item2_name}')


from typing import Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from pyexlatex.presentation.beamer.overlay.overlay import Overlay
from pyexlatex.models.item import NoOptionsNoContentsItem


class BlindText(NoOptionsNoContentsItem):
    """
    Produces lorem ipsum dolor... etc. example text
    """
    name = 'blindtext'

    def __init__(self, overlay: Optional['Overlay'] = None):
        self.init_data()
        self.data.packages.append('blindtext')
        super().__init__(self.name, overlay=overlay)



from typing import Optional, List, Union
from pandas.core.indexes.base import Index as PandasIndex
from itertools import zip_longest
import warnings

from pyexlatex.logic.tools import _max_len_or_zero
from mixins.repr import ReprMixin
from pyexlatex.table.models.labels.collection import LabelCollection
from pyexlatex.table.models.labels.row import LabelRow
from pyexlatex.table.models.table.section import TableSection
from pyexlatex.table.models.texgen.lines import TableLineSegment, TableLineOfSegments


class LabelTable(TableSection, ReprMixin):
    """
    A set of label rows to apply to a DataTable.

    Example:

    >>>import pyexlatex.table as lt
    >>>data_table = lt.DataTable.from_df(some_df)
    >>>col_label_table = lt.LabelTable.from_list_of_lists([['One header'],['More','Headers']])
    >>>row_label_table = lt.LabelTable.from_list_of_lists([['Some','Row','Labels']]).T
    >>>data_table.column_labels = col_label_table
    """
    repr_cols = ['label_collections']

    def __init__(self, label_collections: List[LabelCollection], break_size_adjustment: Optional[str] = None):
        self.label_collections = label_collections
        self.break_size_adjustment = break_size_adjustment

    def __add__(self, other):
        # Import here to avoid circular imports
        from pyexlatex.table.models.data.table import DataTable
        from pyexlatex.table.models.spacing.columntable import ColumnPadTable

        # Return a DataTable if just adding labels to an existing DataTable
        if isinstance(other, DataTable) and not other.row_labels:
            values_table = other.values_table
            column_labels = other.column_labels
            row_labels = self

            return DataTable(
                values_table=values_table,
                column_labels=column_labels,
                row_labels=row_labels
            )
        if isinstance(other, ColumnPadTable):
            self.pad(self.num_columns + other.width)
            return self
        else:
            return super().__add__(other)

    def __iter__(self):
        for collection in self.label_collections:
            yield collection

    def __getitem__(self, item):
        return self.label_collections[item]

    def __contains__(self, item):
        return item in self.label_collections

    @property
    def length(self):
        return len(self.label_collections)

    @property
    def rows(self):
        try:
            return self._rows
        except AttributeError:
            self._rows = self._create_rows()

        return self._rows

    def _create_rows(self):
        len_rows = _max_len_or_zero(self.label_collections)

        rows = []
        label_collection: LabelCollection
        for label_collection in self.label_collections:
            rows.append(
                LabelRow(label_collection)
            )
            if label_collection.underlines is not None:
                rows.append(
                    TableLineOfSegments.from_list_of_ints(label_collection.underlines, num_columns=len(label_collection))
                )

        return rows

    # Following property/setter exist to recreate rows if user overrides labels

    @property
    def label_collections(self):
        return self._label_collections

    @label_collections.setter
    def label_collections(self, label_collections: List[LabelCollection]):
        self._label_collections = label_collections
        self._recreate_rows_if_created()

    @classmethod
    def from_list_of_lists(cls, list_of_lists: List[List[str]]):
        label_collections = []
        for label_list in list_of_lists:
            label_collections.append(
                LabelCollection.from_str_list(label_list)
            )

        return cls(label_collections)

    @classmethod
    def from_df_index(cls, df_columns: PandasIndex):
        column_list = [col for col in df_columns]

        return cls.from_list_of_lists([column_list])

    @property
    def is_empty(self):
        return len(self.label_collections) == 0

    def remove(self, item):
        self.label_collections.remove(item)

    @property
    def T(self):
        return LabelTable(list(map(LabelCollection, zip_longest(*self.label_collections))))

    def matches(self, other):
        """
        Note: use to compare strings inside LabelTables. Use regular equality comparison to compare objects
        :param other:
        :return:
        """
        if not isinstance(other, LabelTable):
            warnings.warn(f'LabelTable.matches() called on type {type(other)}. Will always return False')
            return False

        max_rows = max(len(self.label_collections), len(other.label_collections))

        # Go through all label collections, calling LabelCollection.matches, and returning False if any don't match
        for n_row in range(max_rows):
            if not self.label_collections[n_row].matches(other.label_collections[n_row]):
                return False

        # made it through loop, so all were matching
        return True

    def contains(self, item):
        """
        Note: use to compare strings inside LabelCollections. Use label_collection in LabelTable to compare objects
        :param item:
        :return:
        """

        if not isinstance(item, (str, list, LabelCollection)):
            warnings.warn(f'LabelTable.contains() called on type {type(item)}. Will always return False')
            return False

        item: LabelCollection = LabelCollection.parse_unknown_type(item)

        label_collection: LabelCollection
        for label_collection in self:
            if label_collection.matches(item):
                return True # as soon as one match, return true

        # went through all label collections without a match. no match
        return False

    def begins_with(self, item: Union[str, list, LabelCollection]):
        if not isinstance(item, (str, list, LabelCollection)):
            warnings.warn(f'LabelTable.begins_with() called on type {type(item)}. Will always return False')
            return False

        label: LabelCollection = LabelCollection.parse_unknown_type(item)
        begin_label_collection = LabelCollection([self.label_collections[0][0]])
        return begin_label_collection.matches(label)

    def pad(self, length: int, direction='right'):
        """
        Expand table out to the right or left with blanks, until it is length passed (apply to every row)
        :param length:
        :param direction:
        :return:
        """
        [collection.pad(length=length, direction=direction) for collection in self.label_collections]

    def append(self, item):
        item: LabelCollection = LabelCollection.parse_unknown_type(item)

        # add rather than append directly to activate setter
        self.label_collections = self.label_collections + [item]

    def insert(self, item, index):
        item: LabelCollection = LabelCollection.parse_unknown_type(item)

        # add rather than append directly to activate setter
        self.label_collections.insert(item, index)


import pyexlatex.resume as lr
import pyexlatex as pl
from tests.base import GENERATED_FILES_DIR
from tests.utils.pdf import compare_pdfs_in_generated_vs_input_by_name

JOB_CONTENTS = pl.UnorderedList(['I did', 'some things', 'today'])
JOB = lr.Employment(JOB_CONTENTS, 'Company Name', '2015 - Present', 'Job Title', 'Location, Here', extra_contents='Extra contents')
EDUCATION = lr.Education('Some School', 'Location, Here', 'Bachelor of Science in Something', 'May 2020', '4.0/4.0')
PUBLICATION = lr.Publication(
    'Publication Title',
    ['Coauthor One', 'Coauthor Two'],
    journal_info='Journal of Awesome',
    href='https://www.example.com/',
    extra_info='This was a really neat paper',
    description='A description of the paper'
)
REFERENCE = lr.Reference('My Reference', title_lines=['Reference Title', 'Can be Multiple Lines'],
                         company='Organization', contact_lines=['Address', 'Phone', 'Whatever contact'],
                         email='abc@123.com')
AWARD = lr.Award('Cool Award', 'May 2020', 'Some extra info')
CONTACT_LINES = [
    ['Nick DeRobertis'],
    ['Location, here'],
    ['Contact 1', 'Contact 2']
]

def test_basic_resume():
    contents = [
        pl.Section(
            [
                'An overview'
            ],
            title='Normal section'
        ),
        lr.SpacedSection(
            [
                JOB,
                EDUCATION,
                PUBLICATION,
                AWARD,
                REFERENCE
            ],
            title='Spaced Section'
        )
    ]
    doc = lr.Resume(
        contents,
        contact_lines=CONTACT_LINES
    )
    assert str(doc) == '\\documentclass[11pt]{resume}\n\\newenvironment{absolutelynopagebreak}{\\par\\nobreak\\vfil\\penalty0\\vfilneg\\vtop\\bgroup}{\\par\\xdef\\tpd{\\the\\prevdepth}\\egroup\\prevdepth=\\tpd}\n\\usepackage[left=0.75in,top=0.6in,right=0.75in,bottom=0.6in]{geometry}\n\\usepackage{xcolor}\n\\definecolor{darkblue}{RGB}{50,82,209}\n\\usepackage[hidelinks]{hyperref}\n\\usepackage{ragged2e}\n\\address{Nick DeRobertis}\n\\address{Location, here}\n\\address{Contact 1 \\\\ Contact 2}\n\\begin{document}\n\\begin{section}{Normal section}\nAn overview\n\\end{section}\n\\begin{section}{Spaced Section}\n\\vspace{0.2cm}\n\\begin{absolutelynopagebreak}\n\\begin{employment}{Company Name}{2015 - Present}{Job Title}{Location, Here}\n\\item \\begin{itemize}\n\\item I did\n\\item some things\n\\item today\n\\end{itemize}\nExtra contents\n\\end{employment}\n\\end{absolutelynopagebreak}\n\n\\begin{absolutelynopagebreak}\n\\textbf{Some School}\n (GPA 4.0/4.0)\n\\hfill\n\\textbf{May 2020}\n\\\\\n\\textit{Bachelor of Science in Something}\n\\hfill\nLocation, Here\n\\\\[-8pt]\n\\end{absolutelynopagebreak}\n\n\\begin{absolutelynopagebreak}\n\\hangindent=1cm\n\\href{https://www.example.com/}{\\underline{\\textcolor{darkblue}{Publication Title}}},\n\\textit{Journal of Awesome}\n with \nCoauthor One\n and \nCoauthor Two\nThis was a really neat paper\n\n\\justifying\nA description of the paper\n\\vspace{0.2cm}\n\\vspace{0.2cm}\n\\end{absolutelynopagebreak}\n\n\\begin{absolutelynopagebreak}\n\\textbf{Cool Award}\n (Some extra info)\n\\hfill\nMay 2020\n\\\\[-8pt]\n\\end{absolutelynopagebreak}\n\n\\begin{absolutelynopagebreak}\n\\textsc{\\textbf{My Reference}}\n\\\\\n\\textit{Reference Title}\n\\\\\n\\textit{Can be Multiple Lines}\n\\\\\nOrganization\n\\\\\nAddress\n\\\\\nPhone\n\\\\\nWhatever contact\n\\\\\nabc@123.com\n\\\\\n\\\\[-8pt]\n\\end{absolutelynopagebreak}\n\\vspace{-0.2cm}\n\\end{section}\n\\end{document}'
    name = 'resume basic'
    doc.to_pdf(outfolder=GENERATED_FILES_DIR, outname=name)
    compare_pdfs_in_generated_vs_input_by_name(name)



from typing import Sequence, Tuple, Optional, Union, List, TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pyexlatex.presentation.beamer.overlay.overlay import Overlay
from pyexlatex.graphics.tikz.node.node import Node
from pyexlatex.models.item import ItemBase
from pyexlatex.graphics.tikz.node.position.directions import (
    Above,
    Below,
    Right,
    Left
)
from pyexlatex.models.section.base import TextAreaMixin


class Shape(TextAreaMixin, ItemBase):
    """
    Base class for creating individual shape classes, not intended to be used directly.
    """
    shape_name = '<Do not use Shape directly, set shape_name in subclass>'
    text_node: Optional[Node]

    def __init__(self, shape_options: Optional[List[str]] = None, text_options: Optional[List[str]] = None,
                 offset: Tuple[int, int] = (0, 0),
                 contents: Optional[Any] = None, content_position: str = 'center', content_offset: Optional[float] = None,
                 **kwargs):
        self.content_position = content_position.lower()
        self.content_offset = content_offset
        shape_options = self._get_list_copy_from_list_or_none(shape_options)
        text_options = self._get_list_copy_from_list_or_none(text_options)
        shape_options.extend([
            self.shape_name,
            'draw'
        ])
        self._validate()
        if self.content_position == 'center':
            # pass contents to shape node itself, as shape node is already at center
            shape_contents = contents
            shape_options.extend(text_options)
        else:
            shape_contents = None
        self.shape_node = Node(options=shape_options, location=offset, contents=shape_contents, **kwargs)

        if self.content_position != 'center' and contents is not None:
            # Need to create a second node for placing text other than in center
            text_options.append(self._get_text_placement())
            self.text_node = Node(
                options=text_options,
                location=self._get_text_node_location(),
                contents=contents
            )
        else:
            self.text_node = None

        ItemBase.__init__(self, **kwargs)

        contents = self._get_contents()
        self.add_data_from_content(contents)
        self.contents = self.format_contents(contents)

    def _validate(self):
        self._validate_position()
        self._validate_offset()

    def _validate_position(self):
        allowed_positions = (
            'center',
            'right',
            'left',
            'top',
            'bottom'
        )
        if self.content_position not in allowed_positions:
            raise ValueError(f'could not use content_position {self.content_position}, '
                             f'pass one of {", ".join(allowed_positions)}')

    def _validate_offset(self):
        if self.content_offset is not None and self.content_position == 'center':
            raise ValueError('cannot pass content_offset when content_position is center')

    def _get_text_placement(self) -> Union[Right, Left, Below, Above]:
        """
        Need to place the text beside the line, not on the line. E.g. placing on bottom, we want the text
        above the bottom line, not on the bottom line

        :return:
        """
        if self.content_position == 'right':
            return Left(by=self.content_offset, combined_direction=False)
        elif self.content_position == 'left':
            return Right(by=self.content_offset, combined_direction=False)
        elif self.content_position == 'top':
            return Below(by=self.content_offset, combined_direction=False)
        elif self.content_position == 'bottom':
            return Above(by=self.content_offset, combined_direction=False)
        else:
            raise ValueError(f'could not determine text placement from {self.content_position}')

    def _get_shape_anchor(self) -> str:
        """
        Side of the shape to anchor on for text
        """
        if self.content_position == 'right':
            return 'east'
        elif self.content_position == 'left':
            return 'west'
        elif self.content_position == 'top':
            return 'north'
        elif self.content_position == 'bottom':
            return 'south'
        else:
            raise ValueError(f'could not determine shape anchor from {self.content_position}')

    def _get_text_node_location(self) -> str:
        return f'{self.shape_node.label}.{self._get_shape_anchor()}'

    def _get_contents(self) -> List[Node]:
        possible_contents = [
            self.shape_node,
            self.text_node
        ]
        contents = [content for content in possible_contents if content is not None]
        return contents

    def __str__(self):
        if isinstance(self.contents, str):
            return self.contents
        from pyexlatex.logic.builder import _build
        return _build(self.contents)

    @property
    def label(self) -> Optional[str]:
        return self.shape_node.label



from typing import Dict, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from pyexlatex.presentation.beamer.overlay.overlay import Overlay


def _include_graphics_str(filepath, width=r'\linewidth'):
    return rf'\includegraphics[width={width}]{{{filepath}}}'


def _bracket_modifier_str(modifiers: Optional[str] = None) -> str:
    if modifiers is None:
        return ''

    return f'[{modifiers}]'

def no_options_no_contents_str(item_name: str, overlay: Optional['Overlay'] = None,
                               modifiers: Optional[str] = None) -> str:
    overlay_str = str(overlay) if overlay is not None else ""
    modifiers = modifiers if modifiers is not None else ""
    return rf'\{item_name}{overlay_str}{modifiers}'


def _no_braces_item_str(item_name, contents, overlay: Optional['Overlay'] = None) -> str:
    from pyexlatex.logic.format.contents import format_contents
    overlay_str = str(overlay) if overlay is not None else ""
    return rf'\{item_name}{overlay_str} {format_contents(contents)}'


def _basic_item_str(item_name, contents, modifiers: Optional[str] = None, pre_modifiers: Optional[str] = None,
                    overlay: Optional['Overlay'] = None, format_content: bool = True):
    from pyexlatex.logic.format.contents import format_contents
    pre_modifiers = pre_modifiers if pre_modifiers is not None else ""
    modifiers = modifiers if modifiers is not None else ""
    overlay_str = str(overlay) if overlay is not None else ""

    if format_content:
        contents = format_contents(contents)

    return rf'\{item_name}{overlay_str}{pre_modifiers}{{{contents}}}{modifiers}'


def _multi_option_item_str(item_name, *options, overlay: Optional['Overlay'] = None):
    from pyexlatex.logic.format.contents import format_contents
    overlay_str = str(overlay) if overlay is not None else ""
    options_str = ''.join([f'{{{format_contents(str(option))}}}' for option in options])
    return rf'\{item_name}{overlay_str}{options_str}'


def item_equals_str(item_name, contents):
    return rf'\{item_name}={contents}'


def _cmidrule_str(align, col_str):
    return _multi_option_item_str(rf'cmidrule({align})', col_str)

def _centering_str():
    return r'\centering'

def _toprule_str():
    return r'\toprule'

def _midrule_str():
    return r'\midrule'

def _bottomrule_str():
    return r'\bottomrule'

def _maketitle_str():
    return r'\maketitle'

def _todays_date_str():
    return r'\today'


def _usepackage_str(str_, modifier_str=None):
    if modifier_str:
        full_modifier_str = f'[{modifier_str}]'
    else:
        full_modifier_str = ''
    return rf'\usepackage{full_modifier_str}{{{str_}}}'


def bibtex_str(item_type: str, item_accessor: str, fields: Dict[str, str]) -> str:
    from pyexlatex.logic.builder import _build
    begin_str = f'@{item_type}{{{item_accessor},'
    field_strs = ['    ' + _bibtex_field_str(key, value) for key, value in fields.items()]
    end_str = '}'
    full_str = _build([
        begin_str,
        *field_strs,
        end_str
    ])

    return full_str


def _bibtex_field_str(key: str, value: str) -> str:
    from pyexlatex.logic.format.contents import format_contents
    return f'{key} = {{{format_contents(value)}}},'



from typing import Union, List, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from pyexlatex.presentation.beamer.overlay.overlay import Overlay

from pyexlatex.models.item import SimpleItem, MultiOptionSimpleItem
from pyexlatex.figure.models.graphic import Graphic
from pyexlatex.models.format.raw import Raw


class InlineGraphic(MultiOptionSimpleItem, Graphic):
    """
    A version of Graphic that is meant to be used within text. It centers the text vertically
    when the graphic is taller than the text.
    """
    name = 'vcenteredinclude'
    options: tuple  # type: ignore

    def __init__(self, filepath, width: Union[str, float] = 1.0, cache: bool = True,
                 options: Optional[List[str]] = None,
                 overlay: Optional['Overlay'] = None):
        self.init_data()
        self.add_package('graphicx')
        self.data.packages.append(self._command_definition)

        self._set_path(filepath)
        self.width = width
        self.binaries = None

        self.options_list = self._get_list_copy_from_list_or_none(options)
        self.options_list.append(
            f'width={self.width_str}'
        )

        if cache:
            self._cache(filepath)

        options_str = ','.join(self.options_list)
        self.options = (options_str, self.source_paths[0])
        self.overlay = overlay

    @property
    def _command_definition(self) -> Raw:
        # TODO [#41]: set inline graphic command definition using models
        return Raw(
r"""
\newcommand{\vcenteredinclude}[2]{\begingroup
\setbox0=\hbox{\includegraphics[#1]{#2}}%
\parbox{\wd0}{\box0}\endgroup}
""".strip()
        )


from typing import Union, List

from pyexlatex.models.package import Package

simple_package_strs = [
    'caption',
    'subcaption',
    'graphicx',
    'pdflscape',
    'fancyhdr',
    'lastpage',
    'textcomp'
]
simple_packages = [Package(str_) for str_ in simple_package_strs]

direct_packages: List[Package] = [
    
]

default_packages = simple_packages + direct_packages  # add any packages with options here with the Package class


from pyexlatex.models.section.base import EnvironmentTextArea


class Letter(EnvironmentTextArea):
    name = 'letter'

    def __init__(self, **kwargs):
        super().__init__(name=self.name, **kwargs)



import fitz

from tests.base import INPUT_FILES_DIR, GENERATED_FILES_DIR


def compare_pdfs_in_generated_vs_input_by_name(name: str):
    generated_file = GENERATED_FILES_DIR / f'{name}.pdf'
    input_file = INPUT_FILES_DIR / f'{name}.pdf'
    compare_pdfs(generated_file, input_file)


def compare_pdfs(path_a: str, path_b: str):
    """
    Asserts that two PDFs are equal by loading them from file
    and checking content.

    This is necessary instead of simply comparing bytes because
    the order of certain items in the PDF bytes is not
    guaranteed.

    See https://github.com/rst2pdf/rst2pdf/blob/master/rst2pdf/tests/conftest.py

    :param path_a:
    :param path_b:
    :return:
    """
    pdf_a = fitz.open(path_a)
    pdf_b = fitz.open(path_b)

    # sanity check

    assert pdf_a.isPDF
    assert pdf_b.isPDF

    # compare metadata

    assert _get_metadata(pdf_a) == _get_metadata(pdf_b)

    # compare content

    pages_a = _get_pages(pdf_a)
    pages_b = _get_pages(pdf_b)

    def fuzzy_coord_diff(coord_a, coord_b):
        diff = abs(coord_a - coord_b)
        assert diff / max(coord_a, coord_b) < 0.04  # allow an arbitrary diff

    def fuzzy_string_diff(string_a, string_b):
        words_a = string_a.split()
        words_b = string_a.split()
        assert words_a == words_b

    assert len(pages_a) == len(pages_b)
    for page_a, page_b in zip(pages_a, pages_b):
        assert len(page_a) == len(page_b)
        for block_a, block_b in zip(page_a, page_b):
            # each block has the following format:
            #
            # (x0, y0, x1, y1, "lines in block", block_type, block_no)
            #
            # block_type and block_no should remain unchanged, but it's
            # possible for the blocks to move around the document slightly and
            # the text refold without breaking entirely
            fuzzy_coord_diff(block_a[0], block_b[0])
            fuzzy_coord_diff(block_a[1], block_b[1])
            fuzzy_coord_diff(block_a[2], block_b[2])
            fuzzy_coord_diff(block_a[3], block_b[3])
            fuzzy_string_diff(block_a[4], block_b[4])
            assert block_a[5] == block_b[5]
            assert block_a[6] == block_b[6]


def _get_metadata(pdf):
    metadata = pdf.metadata

    del metadata["creationDate"]
    del metadata["modDate"]

    return metadata


def _get_pages(pdf):
    pages = []

    for page in pdf.pages():
        pages.append(page.getText("blocks"))

    return pages



from pyexlatex.models.documentitem import DocumentItem
from pyexlatex.logic.extract.get import get_attr_from_items_or_collection
from pyexlatex.models.documentsetup import DocumentSetupData


class ContainerItem(DocumentItem):
    """
    Inherit to be able to include LaTeX objects inside LaTeX objects
    """

    def __init__(self, name, content, *args, **kwargs):
        self.add_data_from_content(content)
        super().__init__(name, content, *args, **kwargs)

    def add_data_from_content(self, content):
        self.init_data()
        aggregate_attributes = self.data.attrs
        for attr in aggregate_attributes:
            data_attr = getattr(self.data, attr)
            if attr in self.data.non_unique_attrs:
                unique = False
            else:
                unique = True
            new_values = get_attr_from_items_or_collection(content, attr, unique=unique)
            if new_values is not None:
                data_attr.extend(new_values)

    def init_data(self):
        if not hasattr(self, 'data'):
            self.data = DocumentSetupData()




from pyexlatex.models.section.base import SectionBase
from pyexlatex.models.section.paragraphs import Paragraph


class SubSubSection(SectionBase):
    """
    A part of a subsection, the fourth largest section type.
    """
    name = 'subsubsection'
    next_level_down_class = Paragraph


class SubSection(SectionBase):
    """
    A part of a section, the third largest section type.
    """
    name = 'subsection'
    next_level_down_class = SubSubSection


class Section(SectionBase):
    """
    A section of the document, the second largest section type.
    """
    name = 'section'
    next_level_down_class = SubSection


class Chapter(SectionBase):
    """
    A chapter of the document, the largest section type

    Note: Not supported in all document types. 'report' is one
    type which can support it
    """
    name = 'chapter'
    next_level_down_class = Section



from typing import Optional
from pyexlatex.models.item import ItemBase


class DocumentClassType(ItemBase):

    def __init__(self, name: str, definition: Optional[str] = None):
        self.name = name
        self.definition = definition

        self.init_data()
        if self.definition is not None:
            self.data.filepaths.append(self.class_filename)
            self.data.binaries.append(bytes(self.definition, 'utf8'))

        super().__init__(self.style_source_name)

    @property
    def class_filename(self) -> str:
        return self.name + '.cls'

    @property
    def style_source_name(self) -> str:
        return 'Sources/' + self.name

    def __str__(self):
        return self.name



from pyexlatex.models.item import MultiOptionSimpleItem


class DeclareDelayedFloatFlavor(MultiOptionSimpleItem):
    name = 'DeclareDelayedFloatFlavor'
    equal_attrs = [
        'name',
        'target_environment',
        'treat_as_environment'
    ]

    def __init__(self, target_environment: str, treat_as_environment: str = 'table'):
        self.target_environment = target_environment
        self.treat_as_environment = treat_as_environment
        super().__init__(self.name, target_environment, treat_as_environment)



import conf
major, minor, release = conf.PACKAGE_VERSION_TUPLE
__version__ = f"{major}.{minor}.{release}"
__version_info__ = conf.PACKAGE_VERSION_TUPLE

if __name__ == '__main__':
    print(__version__)


import os

from pyexlatex.logic.output.api.formats import OutputFormats
from pyexlatex.logic.output.to_file import latex_str_to_file
from pyexlatex.logic.output.sources import output_sources_return_tex_input_paths, \
    _write_image_paths_and_binaries_to_folder
from pyexlatex.tools import date_time_move_latex
from pyexlatex.logic.output.fileops import _move_if_exists_and_is_needed
from pyexlatex.texgen.replacements.filename import _latex_valid_basename
from pyexlatex.typing import BytesListOrNone, StrListOrNone


def output_document_and_move(document, outfolder, output_format: OutputFormats = OutputFormats.PDF,
                             image_paths: StrListOrNone = None, outname='figure', as_document=True,
                             move_folder_name='Figures', image_binaries: BytesListOrNone = None,
                             run_bibtex: bool = False, date_time_move: bool = False):

    # Create tex file
    outname_tex = outname + '.tex'
    outpath_tex = os.path.abspath(os.path.join(outfolder, outname_tex))
    with open(outpath_tex, 'w', encoding='utf8') as f:
        f.write(str(document))

    tex_inputs = output_sources_return_tex_input_paths(
        outfolder, image_paths=image_paths, image_binaries=image_binaries
    )

    if as_document:
        outname_final = outname + f'.{output_format.value}'
        outpath_final = os.path.abspath(os.path.join(outfolder, outname_final))
        latex_str_to_file(
            str(document), outpath_final, output_format=output_format, texinputs=tex_inputs, run_bibtex=run_bibtex
        )
    if not date_time_move:
        return

    # Handle date/time move functionality
    new_outfolder = date_time_move_latex(outname, outfolder, folder_name=move_folder_name) #move table into appropriate date/number folder

    if image_paths and new_outfolder:
        # Copy second time to move pictures along with pdf
        sources_tempfolder = os.path.join(outfolder, 'Sources')
        sources_outfolder = os.path.join(new_outfolder, 'Sources')
        if not os.path.exists(sources_outfolder):
            os.makedirs(sources_outfolder)
        if image_binaries:
            _write_image_paths_and_binaries_to_folder(sources_outfolder, image_paths, image_binaries)
        else:
            [_move_if_exists_and_is_needed(
                os.path.join(sources_tempfolder, _latex_valid_basename(filepath)),
                os.path.join(sources_outfolder, _latex_valid_basename(filepath))
             )
            for filepath in image_paths]



from pyexlatex.models.item import SimpleItem
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.typing import PyexlatexItems


class Enclosures(TextAreaMixin, SimpleItem):
    name = 'encl'

    def __init__(self, encl: PyexlatexItems):
        self.encl = encl
        super().__init__(self.name, self.encl)



import pyexlatex as pl
from tests.base import EXAMPLE_IMAGE_PATH, GENERATED_FILES_DIR, INPUT_FILES_DIR
from tests.utils.pdf import compare_pdfs

EXPECT_GRAPHIC = '\\vcenteredinclude{width=0.1\\textwidth}{Sources/nd-logo.png}'
EXPECT_DEFINITION = r"""
\newcommand{\vcenteredinclude}[2]{\begingroup
\setbox0=\hbox{\includegraphics[#1]{#2}}%
\parbox{\wd0}{\box0}\endgroup}
    """.strip()


def test_inline_graphic():
    ig = pl.InlineGraphic(str(EXAMPLE_IMAGE_PATH), width=0.1)
    assert str(ig) == EXPECT_GRAPHIC


def test_inline_graphic_in_document():
    ig = pl.InlineGraphic(str(EXAMPLE_IMAGE_PATH), width=0.1)
    ig2 = pl.InlineGraphic(str(EXAMPLE_IMAGE_PATH), width=0.1)
    contents = ['Some inline text before', ig, 'and after and then wrapping onto the next line so that '
                                               'I can make sure that it is working properly in the case '
                                               'that it is used in a real document', ig2]
    doc = pl.Document(contents)
    assert EXPECT_DEFINITION in str(doc)
    assert EXPECT_GRAPHIC in str(doc)
    doc.to_pdf(GENERATED_FILES_DIR, outname='inline graphic document')
    compare_pdfs(INPUT_FILES_DIR / 'inline graphic document.pdf', GENERATED_FILES_DIR / 'inline graphic document.pdf')



from typing import Optional, Sequence, Union
from pyexlatex.presentation.beamer.frame.frame import Frame
from pyexlatex.models.title.title import Title
from pyexlatex.models.title.subtitle import Subtitle
from pyexlatex.models.credits.author import Author
from pyexlatex.models.date import Date
from pyexlatex.models.title.framepage import MakeFrameTitle


class TitleFrame(Frame):

    def __init__(self, title: Optional[str] = None, authors: Optional[Union[str, Sequence[str]]] = None, date: Optional[str] = None,
                 short_title: Optional[str] = None, subtitle: Optional[str] = None, short_author: Optional[str] = None,
                 institutions: Optional[Sequence[Sequence[str]]] = None, short_institution: Optional[str] = None,
                 **kwargs):
        pre_env_contents = [
            Title(title, short_title=short_title) if title is not None else None,
            Subtitle(subtitle) if subtitle is not None else None,
            Author(
                authors,
                institutions=institutions,
                short_institution=short_institution,
                short_author=short_author
            ) if authors is not None else None,
            Date(date) if date is not None else Date()
        ]

        self.pre_env_content = [content for content in pre_env_contents if content is not None]
        self.add_data_from_content(self.pre_env_content)
        from pyexlatex.logic.builder import _build
        pre_env_contents = _build(self.pre_env_content)

        super().__init__(MakeFrameTitle(), label='title-frame', pre_env_contents=pre_env_contents, **kwargs)


def should_create_title_frame(title: str = None, authors: Optional[Union[str, Sequence[str]]] = None, date: str = None,
                              subtitle: Optional[str] = None, institutions: Optional[Sequence[Sequence[str]]] = None):
    return any([
        title is not None,
        authors is not None,
        date is not None,
        subtitle is not None,
        institutions is not None
    ])


from typing import Optional
from pyexlatex.models.template import Template
import pyexlatex as pl


class Education(Template):
    SPACE_BETWEEN_ADJUSTMENT = '-8pt'

    def __init__(self, school: str, school_location: str, degree: str, expected: str, gpa: Optional[str] = None,
                 prevent_page_break: bool = True):
        self.school = school
        self.school_location = school_location
        self.degree = degree
        self.expected = expected
        self.gpa = gpa
        self.prevent_page_break = prevent_page_break

        self.contents = self._get_contents()
        super().__init__()

    def _get_contents(self):
        possible_items = [
            pl.Bold(self.school),
            f' (GPA {self.gpa})' if self.gpa is not None else None,
            pl.HFill(),
            pl.Bold(self.expected),
            pl.OutputLineBreak(),
            pl.Italics(self.degree),
            pl.HFill(),
            self.school_location,
            pl.OutputLineBreak(size_adjustment=self.SPACE_BETWEEN_ADJUSTMENT)
        ]
        items = [item for item in possible_items if item is not None]
        if self.prevent_page_break:
            items = pl.NoPageBreak(items)
        return items



from pyexlatex.models.references.bibtex.style.builtin_styles import BUILTIN_STYLES
from pyexlatex.models.references.bibtex.style.custom import CUSTOM_STYLES
from pyexlatex.models.references.bibtex.style.style import Style

class StyleManager:

    def __init__(self):
        pass

    def get_style_by_name(self, name: str) -> Style:
        if self.is_builtin_style(name):
            return BUILTIN_STYLES[name]
        
        if self.is_custom_style(name):
            return CUSTOM_STYLES[name]

        raise NoSuchStyleException

    def is_custom_style(self, name: str) -> bool:
        return name in CUSTOM_STYLES

    def is_builtin_style(self, name: str) -> bool:
        return name in BUILTIN_STYLES


class NoSuchStyleException(Exception):
    pass



import pathlib
from typing import Sequence

from pyexlatex.logic.output.api.builders.base import BaseBuilder


class HTLatexBuilder(BaseBuilder):
    """A simple htlatex-based builder for LaTeX files to generate HTML from LaTeX.

    Builds LaTeX files by copying them to a temporary directly and running
    ``htlatex`` until the associated ``.aux`` file stops changing.

    :param executable: The path to the ``htlatex`` binary (will looked up on
                    ``$PATH``).
    :param max_runs: An integer providing an upper limit on the amount of times
                     ``htlatex`` can be rerun before an exception is thrown.
    """
    output_extension = 'html'
    default_executable = 'make4ht'
    pre_file_output_args: Sequence[str] = ('--lua', '--shell-escape', '--utf8')
    post_file_output_args: Sequence[str] = tuple()

    def _pre_compile(self, temp_dir: str, base_file_name: str):
        config_contents = """
settings_add{ tex4ht_sty_par =  "html,css-in" }
Make:htlatex()
Make:htlatex()
Make:htlatex()
        """.strip()
        out_path = pathlib.Path(temp_dir) / f'{base_file_name}.mk4'
        out_path.write_text(config_contents)


from pyexlatex import Document
from pyexlatex.models.format.captionsetup import CaptionSetup


def test_caption_setup_no_options():
    content = CaptionSetup()
    assert str(content) == '\\captionsetup[figure]{}'


def test_caption_setup_options():
    content = CaptionSetup(
        'table', relative_sizes={'font': -1, 'labelfont': 1}, options=['format=plain', 'indentation=.5cm']
    )
    assert str(content) == '\\captionsetup[table]{format=plain,indentation=.5cm,font=small,labelfont=large}'


def test_caption_in_document():
    content = CaptionSetup()
    doc = Document(['woo'], extra_pre_env_contents=[content])
    assert '\\captionsetup[figure]{}' in str(doc)
    assert '\\usepackage{caption}' in str(doc)




from typing import Dict, Sequence, Optional

from pyexlatex.models.sizes.textsizes import TextSize
from pyexlatex.models.item import SimpleItem


class CaptionSetup(SimpleItem):
    """
    Control formatting of captions for tables and figures
    """
    name = 'captionsetup'

    def __init__(self, target: str = 'figure', relative_sizes: Optional[Dict[str, int]] = None,
                 options: Optional[Sequence[str]] = None):
        self.init_data()
        self.add_package('caption')

        if relative_sizes is None:
            relative_sizes = {}

        if options is None:
            options = []
        else:
            options = list(options)

        size_options = {name: TextSize(size).name for name, size in relative_sizes.items()}
        size_options_str = SimpleItem._dict_to_options_str(size_options)
        options_str = ','.join([*options, size_options_str])

        target_str = self._wrap_with_bracket(target)

        super().__init__(self.name, options_str, pre_modifiers=target_str)



from tempdir import TempDir

from pyexlatex.logic.output.api.main import latex_str_to_obj
from pyexlatex.logic.output.api.formats import OutputFormats
from pyexlatex.logic.output.sources import output_sources_return_tex_input_paths
from pyexlatex.typing import StrListOrNone, BytesListOrNone


def latex_str_to_file(latex_str: str, filepath: str, output_format: OutputFormats = OutputFormats.PDF,
                      texinputs: StrListOrNone = None,
                      run_bibtex: bool = False):
    obj = latex_str_to_obj(latex_str, output_format=output_format, texinputs=texinputs, run_bibtex=run_bibtex)
    obj.save_to(filepath)
    return obj


def latex_str_to_file_obj_with_sources(latex_str: str, output_format: OutputFormats = OutputFormats.PDF,
                                       image_paths: StrListOrNone = None,
                                       image_binaries: BytesListOrNone = None, run_bibtex: bool = False):
    with TempDir() as tmpdir:
        tex_inputs = output_sources_return_tex_input_paths(
            tmpdir, image_paths=image_paths, image_binaries=image_binaries
        )
        obj = latex_str_to_obj(latex_str, output_format=output_format, texinputs=tex_inputs, run_bibtex=run_bibtex)

    return obj


from sympy import Eq, Expr, Symbol
from sympy import latex


def inline(eq: str) -> str:
    return f'${eq}$'


def offset(eq: str) -> str:
    return f'\\begin{{equation}}\n\t{eq}\n\\end{{equation}}'


def offset_no_numbering(eq: str) -> str:
    return f'\n$${eq}$$\n'


def latex_partial(expr: Expr, eq_symbol: Symbol, wrt_symbol: Symbol, offset: bool = False) -> str:

    begin_str = '$'
    end_str = '$'
    if offset:
        begin_str = '\\begin{equation}\n\t'
        end_str = '\n\\end{equation}'
    return rf'{begin_str}\frac{{\partial {latex(eq_symbol)}}}{{\partial {latex(wrt_symbol)}}} = ' \
        rf'{latex(expr)}{end_str}'



from typing import Sequence
from pyexlatex.models.section.base import TextAreaBase
from pyexlatex.models.package import Package


class AdjustBox(TextAreaBase):
    name = 'adjustbox'

    def __init__(self, contents, adjust_options: Sequence[str]):
        self.adjust_options = adjust_options
        self.init_data()
        self.data.packages.append(Package('adjustbox'))
        super().__init__(self.name, contents, env_modifiers=self.options_str)

    @property
    def options_str(self) -> str:
        base_str = ', '.join(self.adjust_options)
        result = self._wrap_with_braces(base_str)
        if result is None:  # for type checking, shouldn't actually hit here
            return ''
        return result



from typing import List

import numpy as np

from pyexlatex.table.models.table.section import TableSection
from mixins.repr import ReprMixin


class PanelGrid(ReprMixin):
    repr_cols = ['sections', 'shape']

    def __new__(cls, sections: List[TableSection], shape: tuple=None):
        index_arr = index_array(sections, shape)
        return GridShape.__new__(GridShape, index_arr)


class GridShape(np.ndarray):

    def __new__(cls, input_array, info=None):
        # Input array is an already formed ndarray instance
        # We first cast to be our class type
        obj = np.asarray(input_array).view(cls)
        # add the new attribute to the created instance
        obj.info = info
        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # see InfoArray.__array_finalize__ for comments
        if obj is None: return
        self.info = getattr(obj, 'info', None)

    @property
    def data_has_row_labels(self) -> bool:
        from pyexlatex.table.models.data.table import DataTable
        for section_arr in np.nditer(self, flags=('refs_ok',)):
            section = section_arr.item()
            if isinstance(section, DataTable):
                if section.row_labels is not None and len(section.row_labels.label_collections) > 0:
                    return True
        return False

    @property
    def cell_widths(self) -> List[int]:
        n_cols = self.shape[1]
        widths: List[int] = []
        for row in self:
            width = 0
            for i in range(n_cols):
                section: TableSection = row[i]
                width += section.cell_width
            widths.append(width)
        return widths

    @property
    def cell_width(self) -> int:
        return max(self.cell_widths)


def index_array(sections: List[TableSection], shape: tuple=None):
    if shape is None:
    # default is one column, as many rows as sections
        if sections:
            shape = (len(sections), 1)
        else:
            raise ValueError(f'must pass tuple of shape or sections, got shape: {shape}, sections: {sections}')

    return np.array(sections).reshape(shape)




import os
from typing import Union, List, Dict, Any, TYPE_CHECKING, Optional

from pyexlatex.typing import PyexlatexItems

if TYPE_CHECKING:
    from matplotlib.pyplot import Axes, Figure as PltFigure

from pyexlatex.figure.models.subfigure import Subfigure, Graphic
from pyexlatex.models.item import Item
from pyexlatex.models.caption import Caption
from pyexlatex.models.label import Label
from pyexlatex.logic.builder import build_figure_content
from pyexlatex.texgen.replacements.filename import latex_filename_replacements

from pyexlatex.models.commands.newenvironment import NewEnvironment
from pyexlatex.models.commands.begin import Begin
from pyexlatex.models.commands.end import End
from pyexlatex.models.environment import Environment
from pyexlatex.models.containeritem import ContainerItem

SubfigureOrGraphic = Union[Subfigure, Graphic]
SubfiguresOrGraphics = List[SubfigureOrGraphic]
PltFigureOrAxes = Union['Axes', 'PltFigure']
PltFigureOrAxesNameDict = Dict[str, PltFigureOrAxes]


class Figure(ContainerItem, Item):
    """
    Used for creating latex figures from images. Currently the main usage is the Figure class created with the method
    Figure.from_dict_of_names_and_filepaths. Pass a dictionary where the keys are names for subfigures and the values
    are filepaths where the image for the subfigure is located.

    """
    name = 'figure'

    def __init__(self, subfigures: SubfiguresOrGraphics, caption: Optional[PyexlatexItems] = None,
                 label: Optional[str] = None, centering: bool = True, position_str: Optional[str] = None,
                 landscape: bool = False):
        self.subfigures = subfigures
        self.caption = Caption(caption) if caption else None
        self.label = Label(label) if label else None
        self.centering = centering
        self.landscape = landscape

        self._remove_subfigure_elevate_contents_to_figure_if_single_subfigure()

        self.add_data_from_content(self.subfigures)

        content = build_figure_content(
            self.subfigures,
            caption=self.caption,
            label=self.label,
            centering=self.centering,
            position_str=position_str
        )

        super().__init__(self.name, content)

        if landscape:
            lfigure_def = NewEnvironment(
                'lfigure',
                Begin('landscape') + Begin('figure'),
                End('figure') + End('landscape')
            )
            self.data.begin_document_items.append(lfigure_def)
            self.env = Environment('lfigure')

    def __repr__(self):
        return f'<Figure(subfigures={self.subfigures}, caption={self.caption})>'

    def __iter__(self):
        for subfigure in self.subfigures:
            yield subfigure

    def __getitem__(self, item):
        return self.subfigures[item]

    def as_document(self, landscape=False):
        from pyexlatex.models.document import Document
        from pyexlatex.figure.packages import default_packages

        return Document(self, default_packages, landscape=landscape)

    def to_pdf_and_move(self, as_document=True, outfolder: str=None, outname: str=None,
                        landscape=False):
        from pyexlatex.logic.output.main import output_document_and_move
        from pyexlatex.models.document import Document

        to_output: Union[Figure, Document]
        to_output = self  # Figure
        if as_document:
            to_output = self.as_document(  # Document
                landscape=landscape if self.landscape == False else False  # don't apply landscape twice
            )

        if outfolder is None:
            outfolder = '.'
        if outname is None:
            outname = 'figure'
        else:
            outname = latex_filename_replacements(outname)

        output_document_and_move(
            to_output,
            outfolder,
            image_paths=self.data.filepaths,
            outname=outname,
            as_document=as_document,
            image_binaries=self.data.binaries
        )

    @classmethod
    def from_dict_of_names_and_filepaths(cls, filepath_name_dict: dict, figure_name: str=None,
                                         position_str_name_dict: dict=None, **kwargs):
        """
        Create a Figure from a dictionary where keys are names of subfigures and values are file paths

        :param filepath_name_dict: dictionary where keys are names of subfigures and values
            are the filepaths to the images for those subfigures.
        :param figure_name: name for overall figure
        :param position_str_name_dict: dictionary where keys are names of subfigures and values
            are the position strs for those figures, e.g. r'[t]{0.45\linewidth}'
        :param kwargs: kwargs for Figure
        :return: Figure
        """

        # TODO [#3]: for Figure, add possibility of passing grid shape rather than actual position str

        if position_str_name_dict is None:
            position_str_name_dict = {}

        subfigures = []
        for name, filepath in filepath_name_dict.items():
            subfigures.append(
                Subfigure(
                    filepath,
                    caption=name,
                    position_str=position_str_name_dict[name] if name in position_str_name_dict else r'[t]{0.45\linewidth}'
                )
            )

        return cls(
            subfigures,
            caption=figure_name,
            **kwargs
        )

    @classmethod
    def from_dict_of_names_and_plt_figures(cls, plt_fig_name_dict: PltFigureOrAxesNameDict, sources_outfolder: str,
                                           source_filetype: str = 'pdf',
                                           figure_name: str=None,
                                           position_str_name_dict: dict=None, **kwargs):
        """
        Create a Figure from a dictionary where keys are names of subfigures and values are matplotlib Figures
        or Axes

        :param plt_fig_name_dict: Key is display name in output figure, value is matplotlib axes or figure
        :param sources_outfolder: folder to output individual matplotlib figures
        :param source_filetype: Filetype for individual plt figures. The default is pdf. Use png or another
            image type if outputting complicated figures or performance may be affected when viewing the pdf.
        :param figure_name: name for overall figure
        :param position_str_name_dict: dictionary where keys are names of subfigures and values
            are the position strs for those figures, e.g. r'[t]{0.45\linewidth}'
        :param kwargs: kwargs for Figure
        :return: Figure
        """

        filepath_name_dict = {}  # store outputted filepaths of sources to pass to from_dict_of_names_and_filepaths
        for name, plt_figure_or_axes in plt_fig_name_dict.items():
            plt_figure = _get_plt_figure_from_axes_or_figure(plt_figure_or_axes)
            outpath = os.path.join(sources_outfolder, f'{latex_filename_replacements(name)}.{source_filetype}')
            plt_figure.savefig(outpath)
            filepath_name_dict[name] = outpath

        return cls.from_dict_of_names_and_filepaths(
            filepath_name_dict,
            figure_name=figure_name,
            position_str_name_dict=position_str_name_dict,
            **kwargs
        )



    def _remove_subfigure_elevate_contents_to_figure_if_single_subfigure(self):
        """
        If there is a single subfigure, leaving in subfigure format results in very small output. Must strip
        out subfigure layer, leaving only a figure with a graphic, then it can fill the page.
        :return:
        """
        if len(self.subfigures) != 1:
            return

        self.subfigures: List[Union[Subfigure, Graphic]]

        if hasattr(self.subfigures[0], 'graphic'):
            # got subfigure
            item_is_subfigure = True
            orig_graphic = self.subfigures[0].graphic
        else:
            # got graphic
            item_is_subfigure = False
            orig_graphic = self.subfigures[0]

        orig_width = self.subfigures[0].width

        # Elevate caption of sub-figure if there is no figure caption
        if self.caption is None and item_is_subfigure:
            self.caption = self.subfigures[0].caption

        # update width to original width
        graphic = Graphic(orig_graphic.filepaths[0], width=orig_width)

        # need to turn off centering to cover whole page
        self.centering = False

        self.subfigures = [graphic]

    def to_graphic_list(self) -> List[Graphic]:
        graphics: List[Graphic] = []
        for subfig_or_graphic in self.subfigures:
            if isinstance(subfig_or_graphic, Graphic):
                graphics.append(subfig_or_graphic)
            elif isinstance(subfig_or_graphic, Subfigure):
                graphics.append(subfig_or_graphic.graphic)
            else:
                raise ValueError(f'got other than Subfigure or Graphic in subfigures: '
                                 f'{subfig_or_graphic} of type {type(subfig_or_graphic)}')
        return graphics

def _get_plt_figure_from_axes_or_figure(plt_axes_or_fig: PltFigureOrAxes) -> 'PltFigure':
    # Both axes and figure have the get_figure method, however for the figure, it will return None
    possible_figure = plt_axes_or_fig.get_figure()
    if possible_figure is None:
        return plt_axes_or_fig  # had figure already
    else:
        return possible_figure # extracted figure from axes






import os
from typing import Optional, List

from pyexlatex.logic.output.fileops import _copy_if_needed
from pyexlatex.texgen.replacements.filename import _latex_valid_basename
from pyexlatex.typing import StrListOrNone, BytesListOrNone, StrList, BytesList


def output_sources_return_tex_input_paths(outfolder: str, image_paths: StrListOrNone = None,
                                          image_binaries: BytesListOrNone = None) -> Optional[List[str]]:
    tex_inputs: Optional[List[str]]
    if image_paths:
        # Copy first time for creation of pdf
        sources_tempfolder = os.path.join(outfolder, 'Sources')
        tex_inputs = [os.path.abspath(outfolder), os.path.abspath(sources_tempfolder), '.']
        if not os.path.exists(sources_tempfolder):
            os.makedirs(sources_tempfolder)
        if image_binaries:
            _write_image_paths_and_binaries_to_folder(sources_tempfolder, image_paths, image_binaries)
        else:
            [_copy_if_needed(filepath, os.path.join(sources_tempfolder, _latex_valid_basename(filepath)))
             for filepath in image_paths]
    else:
        tex_inputs = None

    return tex_inputs


def _write_image_paths_and_binaries_to_folder(folder: str, image_paths: StrList, image_binaries: BytesList):
    if len(image_binaries) != len(image_paths):
        raise ValueError('must have equal image_binaries and image_path lengths if image_binaries are passed')
    for filepath, binary in zip(image_paths, image_binaries):
        image_outpath = os.path.join(folder, _latex_valid_basename(filepath))
        with open(image_outpath, 'wb') as f:
            f.write(binary)


from typing import List, Dict, Union
import warnings
from latex.exc import LatexBuildError
from pyexlatex.logic.output.errors.models.error import LatexError


class LatexErrorCollection:

    def __init__(self, latex_build_error: LatexBuildError):
        self.exc = latex_build_error
        self.errors = self._parse_errors()

    def __getitem__(self, index):
        return self.errors[index]

    def __iter__(self):
        return iter(self.errors)

    def _parse_errors(self) -> List[LatexError]:
        error_dict_list = get_errors_dict_list(self.exc)
        errors = []
        for error_dict in error_dict_list:
            arg_dict = error_dict.copy()
            arg_dict.pop('error')
            errors.append(LatexError(**arg_dict))  # type: ignore
        if not errors:
            warnings.warn(f'got LatexBuildError with no errors attached: {self.exc}')
        return errors

def get_errors_dict_list(exc: LatexBuildError) -> List[Dict[str, Union[str, int, List[str]]]]:
    try:
        return exc.get_errors()
    except AttributeError as e:
        if "'NoneType' object has no attribute 'splitlines'" in str(e):
            return []
        raise e


from pyexlatex.models.template import Template


class Group(Template):
    """
    Groups together latex items so that "global" commands will only affect the items in a group
    """

    def __init__(self, contents):
        if not isinstance(contents, (list, tuple)):
            contents = [contents]
        contents.insert(0, '{')
        contents.append('}')
        self.contents = contents
        super().__init__()



from pyexlatex.models.item import SimpleItem
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.typing import PyexlatexItems


class Signature(TextAreaMixin, SimpleItem):
    name = 'signature'

    def __init__(self, signer: PyexlatexItems):
        self.signer = signer
        super().__init__(self.name, self.signer)



from pyexlatex.models.environment import Environment

class TableEnvironment(Environment):
    name = 'table'

    def __init__(self):
        super().__init__(name=self.name)

class ThreePartTableEnvironment(Environment):
    name = 'threeparttable'

    def __init__(self):
        super().__init__(name=self.name)

class TabularEnvironment(Environment):
    name = 'tabular'

    def __init__(self):
        super().__init__(name=self.name)



from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.models.item import SimpleItem


class Underline(TextAreaMixin, SimpleItem):
    """
    Adds an underline to text.
    """
    name = 'underline'

    def __init__(self, contents, **kwargs):
        super().__init__(self.name, contents, **kwargs)



from pyexlatex.models.item import SimpleItem


class Label(SimpleItem):
    """
    A label which can later be referred to by Ref
    """
    name = 'label'

    def __init__(self, contents):
        super().__init__(self.name, contents)


from typing import List, Optional, Dict, Any
import warnings

from pyexlatex.logic.output.api.formats import OutputFormats
from pyexlatex.logic.output.errors.exc import (
    LatexException
)
from pyexlatex.logic.output.api.exc_handler.prepend.main import handle_prepend_exceptions
from pyexlatex.logic.output.api.exc_handler.prepend.typing import PrependKwargsDict, PrependItemsDict


class APIExceptionHandler:

    def __init__(self, exceptions: List[LatexException], orig_exception: Exception, latex_str: str,
                 prepend_kwargs_dict: PrependKwargsDict = None, prepend_items_dict: PrependItemsDict = None,
                 retries_remaining: int = 3, output_format: OutputFormats = OutputFormats.PDF, **latex_kwargs):
        self.exceptions = exceptions
        self.orig_exception = orig_exception
        self.latex_str = latex_str
        self.prepend_kwargs_dict = prepend_kwargs_dict
        self.prepend_items_dict = prepend_items_dict
        self.retries_remaining = retries_remaining
        self.output_format = output_format
        self.latex_kwargs = latex_kwargs

    def handle_exceptions(self):
        from pyexlatex.logic.output.api.main import latex_str_to_obj
        if not self.exceptions:
            # Got LatexBuildError, but could not extract any exceptions from it. Something is going wrong
            # Seems like it might be some intermittent issue, try retrying
            if self.retries_remaining > 0:
                warnings.warn('got empty latex build error. trying to create pdf again')
                return latex_str_to_obj(
                    self.latex_str,
                    output_format=self.output_format,
                    retries_remaining=self.retries_remaining - 1,
                    prepend_items_dict=self.prepend_items_dict,
                    prepend_kwargs_dict=self.prepend_kwargs_dict,
                    **self.latex_kwargs
                )
            raise LatexException(self.orig_exception)
        prepend_items_dict, prepend_kwarg_dict, unhandled_exceptions = handle_prepend_exceptions(
            self.exceptions, self.prepend_kwargs_dict, self.prepend_items_dict
        )

        # TODO [#10]: handle other exceptions
        #
        # Not actually sure what this is for looking back at the code, but upon closer review
        # perhaps some additional exception handling will be needed.

        if len(unhandled_exceptions) == len(self.exceptions):
            # was not able to handle any exceptions, so retrying would be of no use
            raise LatexException(self.exceptions)

        return latex_str_to_obj(
            self.latex_str,
            self.output_format,
            retries_remaining=self.retries_remaining,
            prepend_items_dict=prepend_items_dict,
            prepend_kwargs_dict=prepend_kwarg_dict,
            **self.latex_kwargs
        )









import pandas as pd

import pyexlatex as pl

from tests.base import GENERATED_FILES_DIR, EXAMPLE_IMAGE_PATH
from tests.utils.pdf import compare_pdfs_in_generated_vs_input_by_name


def test_basic_presentation():
    doc = pl.Presentation(['woo'])
    assert str(doc) == '\\documentclass[11pt]{beamer}\n\\mode\n<presentation>{\\usetheme{Madrid}}\n\\begin{document}\nwoo\n\\end{document}'
    name = 'basic presentation'
    doc.to_pdf(outfolder=GENERATED_FILES_DIR, outname=name)
    compare_pdfs_in_generated_vs_input_by_name(name)


def test_presentation_options():
    doc = pl.Presentation(
        [
            pl.Section(
                [
                    pl.Frame(
                        [
                            'woo'
                        ],
                        title='One-One'
                    ),
                    pl.Frame(
                        [
                            'woo2'
                        ],
                        title='One-Two'
                    )
                ],
                title="One"
            ),
            pl.Section(
                [
                    pl.Frame(
                        [
                            'woo3'
                        ],
                        title='Two-One'
                    ),
                    pl.Frame(
                        [
                            'woo4'
                        ],
                        title='Two-Two'
                    )
                ],
                title="Two"
            )
        ],
        packages=['hyperref'],
        title='Test Presentation',
        authors=['Nick DeRobertis', 'Someone Else'],
        date='2020-10-19',
        short_title='Pres',
        subtitle='A Presentation for Testing Purposes',
        short_author='People',
        institutions=[['University of Florida', 'Line two'], ['Virginia Commonwealth University']],
        short_institution='UF & VCU',
        font_size=15,
        nav_header=True,
        toc_sections=True,
    )
    assert str(doc) == '\\documentclass[15pt]{beamer}\n\\mode\n<presentation>{\\usetheme{Madrid}}\n\\institute[UF \\& VCU]{\\inst{1}\nUniversity of Florida\\\\\nLine two\n\\and\n\\inst{2}\nVirginia Commonwealth University}\n\\setbeamertemplate{headline}{\\begin{beamercolorbox}[ht=2.25ex, dp=3.75ex]{section in head/foot}\n\\insertnavigation{\\paperwidth}\n\\end{beamercolorbox}}\n\\AtBeginSection{\\begin{frame}\n\\frametitle{Table of Contents}\n\\tableofcontents[currentsection]\n\\end{frame}}\n\\begin{document}\n\\title[Pres]{Test Presentation}\n\\subtitle{A Presentation for Testing Purposes}\n\\author[People]{Nick DeRobertis\\inst{1}, Someone Else\\inst{2}}\n\\date{2020-10-19}\n\\begin{frame}\n\\titlepage\n\\label{title-frame}\n\\end{frame}\n\\begin{section}{One}\n\\begin{frame}\n\\frametitle{One-One}\nwoo\n\\end{frame}\n\\begin{frame}\n\\frametitle{One-Two}\nwoo2\n\\end{frame}\n\\end{section}\n\\begin{section}{Two}\n\\begin{frame}\n\\frametitle{Two-One}\nwoo3\n\\end{frame}\n\\begin{frame}\n\\frametitle{Two-Two}\nwoo4\n\\end{frame}\n\\end{section}\n\\end{document}'
    name = 'presentation with options'
    doc.to_pdf(outfolder=GENERATED_FILES_DIR, outname=name)
    compare_pdfs_in_generated_vs_input_by_name(name)


def test_table_in_presentation():
    df = pd.DataFrame(
        [
            (1, 2, 3.546),
            (4, 5, 66546.4),
            (7, 8, 96.54)
        ],
        columns=['a', 'b', 'c']
    )
    table = pl.Table.from_list_of_lists_of_dfs(
        [[df]],
        caption='My Table Title',
        below_text='My below text',
        align='L{5cm}c.',
        mid_rules=False,
    )
    doc = pl.Presentation(
        [
            pl.Section(
                [
                    pl.Frame(
                        [
                            table
                        ],
                        title='Table'
                    ),
                ],
                title="Section"
            ),
        ],
    )
    assert str(doc) == '\\documentclass[11pt]{beamer}\n\\mode\n<presentation>{\\usetheme{Madrid}}\n\\usepackage{threeparttable}\n\\usepackage{booktabs}\n\\usepackage{dcolumn}\n\\newcolumntype{L}[1]{>{\\raggedright\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{C}[1]{>{\\centering\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{R}[1]{>{\\raggedleft\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{.}{D{.}{.}{-1}}\n\\begin{document}\n\\begin{section}{Section}\n\\begin{frame}\n\\frametitle{Table}\n\\begin{table}\n\\centering\n\\begin{threeparttable}\n\\caption{My Table Title}\n\\begin{tabular}{L{5cm}c.}\n\\toprule\na & b & c\\\\\n 1 &  2 &      3.546 \\\\\n 4 &  5 &  66546.400 \\\\\n 7 &  8 &     96.540 \\\\\n\\bottomrule\n\n\\end{tabular}\n\\begin{tablenotes}[para, flushleft]\nMy below text\n\\end{tablenotes}\n\\end{threeparttable}\n\\end{table}\n\\end{frame}\n\\end{section}\n\\end{document}'
    name = 'presentation with table'
    doc.to_pdf(outfolder=GENERATED_FILES_DIR, outname=name)
    compare_pdfs_in_generated_vs_input_by_name(name)


def test_tabular_from_table_in_presentation():
    df = pd.DataFrame(
        [
            (1, 2, 3.546),
            (4, 5, 66546.4),
            (7, 8, 96.54)
        ],
        columns=['a', 'b', 'c']
    )
    table = pl.Table.from_list_of_lists_of_dfs(
        [[df], [df + 10]],
        caption='My Table Title',
        below_text='My below text',
        align='L{5cm}c.'
    )
    doc = pl.Presentation(
        [
            pl.Section(
                [
                    pl.Frame(
                        [
                            table.tex_obj(as_document=False, as_single_tabular=True),
                        ],
                        title='Tabular'
                    ),
                ],
                title="Section"
            ),
        ],
    )
    assert str(doc) == '\\documentclass[11pt]{beamer}\n\\mode\n<presentation>{\\usetheme{Madrid}}\n\\usepackage{booktabs}\n\\usepackage{dcolumn}\n\\newcolumntype{L}[1]{>{\\raggedright\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{C}[1]{>{\\centering\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{R}[1]{>{\\raggedleft\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{.}{D{.}{.}{-1}}\n\\begin{document}\n\\begin{section}{Section}\n\\begin{frame}\n\\frametitle{Tabular}\n\\begin{tabular}{L{5cm}c.}\n\\toprule\na & b & c\\\\\n\\midrule\n 1 &  2 &      3.546 \\\\\n 4 &  5 &  66546.400 \\\\\n 7 &  8 &     96.540 \\\\\n  &   &  \\\\\n 11 &  12 &     13.546 \\\\\n 14 &  15 &  66556.400 \\\\\n 17 &  18 &    106.540 \\\\\n\\bottomrule\n\n\\end{tabular}\n\\end{frame}\n\\end{section}\n\\end{document}'
    name = 'presentation with tabular'
    doc.to_pdf(outfolder=GENERATED_FILES_DIR, outname=name)
    compare_pdfs_in_generated_vs_input_by_name(name)


def test_tabular_list_from_table_in_presentation():
    df = pd.DataFrame(
        [
            (1, 2, 3.546),
            (4, 5, 66546.4),
            (7, 8, 96.54)
        ],
        columns=['a', 'b', 'c']
    )
    table = pl.Table.from_list_of_lists_of_dfs(
        [[df], [df + 10]],
        caption='My Table Title',
        below_text='My below text',
        align='L{5cm}c.'
    )
    frames = [
        pl.Frame(tab, title=f'Tabular {i + 1}')
        for i, tab in enumerate(table.tex_obj(as_document=False, as_panel_tabular_list=True))
    ]

    doc = pl.Presentation(
        [
            pl.Section(
                [
                    frames,
                ],
                title="Section"
            ),
        ],
    )
    assert str(doc) == '\\documentclass[11pt]{beamer}\n\\mode\n<presentation>{\\usetheme{Madrid}}\n\\usepackage{booktabs}\n\\usepackage{dcolumn}\n\\newcolumntype{L}[1]{>{\\raggedright\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{C}[1]{>{\\centering\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{R}[1]{>{\\raggedleft\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{.}{D{.}{.}{-1}}\n\\begin{document}\n\\begin{section}{Section}\n\\begin{frame}\n\\frametitle{Tabular 1}\n\\begin{tabular}{L{5cm}c.}\n\\toprule\na & b & c\\\\\n\\midrule\n 1 &  2 &      3.546 \\\\\n 4 &  5 &  66546.400 \\\\\n 7 &  8 &     96.540 \\\\\n\\bottomrule\n\n\\end{tabular}\n\\end{frame}\n\\begin{frame}\n\\frametitle{Tabular 2}\n\\begin{tabular}{L{5cm}c.}\n\\toprule\na & b & c\\\\\n\\midrule\n 11 &  12 &     13.546 \\\\\n 14 &  15 &  66556.400 \\\\\n 17 &  18 &    106.540 \\\\\n\\bottomrule\n\n\\end{tabular}\n\\end{frame}\n\\end{section}\n\\end{document}'
    name = 'presentation with tabular list'
    doc.to_pdf(outfolder=GENERATED_FILES_DIR, outname=name)
    compare_pdfs_in_generated_vs_input_by_name(name)


def test_figure_in_presentation():
    graphic = pl.Graphic(str(EXAMPLE_IMAGE_PATH), width=0.4)
    fig = pl.Figure([graphic], caption='My Figure')
    doc = pl.Presentation(
        [
            pl.Section(
                [
                    pl.Frame(
                        [
                            fig
                        ],
                        title='Figure'
                    ),
                ],
                title="Section"
            ),
        ],
    )
    assert str(doc) == '\\documentclass[11pt]{beamer}\n\\mode\n<presentation>{\\usetheme{Madrid}}\n\\begin{document}\n\\begin{section}{Section}\n\\begin{frame}\n\\frametitle{Figure}\n\\begin{figure}\n\\includegraphics[width=0.4\\textwidth]{Sources/nd-logo.png}\n\\caption{My Figure}\n\\end{figure}\n\\end{frame}\n\\end{section}\n\\end{document}'
    name = 'presentation with figure'
    doc.to_pdf(outfolder=GENERATED_FILES_DIR, outname=name)
    compare_pdfs_in_generated_vs_input_by_name(name)


def test_graphic_from_figure_in_presentation():
    graphic = pl.Graphic(str(EXAMPLE_IMAGE_PATH), width=0.4)
    fig = pl.Figure([graphic], caption='My Figure')
    doc = pl.Presentation(
        [
            pl.Section(
                [
                    pl.Frame(
                        [
                            fig.to_graphic_list()[0]
                        ],
                        title='Graphic'
                    ),
                ],
                title="Section"
            ),
        ],
    )
    assert str(doc) == '\\documentclass[11pt]{beamer}\n\\mode\n<presentation>{\\usetheme{Madrid}}\n\\begin{document}\n\\begin{section}{Section}\n\\begin{frame}\n\\frametitle{Graphic}\n\\includegraphics[width=0.4\\textwidth]{Sources/nd-logo.png}\n\\end{frame}\n\\end{section}\n\\end{document}'
    name = 'presentation with graphic from figure'
    doc.to_pdf(outfolder=GENERATED_FILES_DIR, outname=name)
    compare_pdfs_in_generated_vs_input_by_name(name)


from copy import deepcopy

import pandas as pd

import pyexlatex as pl
from tests.base import GENERATED_FILES_DIR
from tests.utils.pdf import compare_pdfs_in_generated_vs_input_by_name

EXAMPLE_DF = pd.DataFrame(
    [
        (1, 2, 3),
        (4, 5, 6),
        (7, 8, 9)
    ],
    columns=['a', 'b', 'c']
)
EXAMPLE_DF_WITH_NAMED_INDEX = deepcopy(EXAMPLE_DF)
EXAMPLE_DF_WITH_NAMED_INDEX.index.name = 'index_name'

class TestTable:
    table = pl.Table.from_list_of_lists_of_dfs(
        [[EXAMPLE_DF]],
        caption='My Table Title',
        below_text='My below text',
        mid_rules=False,
    )
    two_panel_table_from_dict_no_index = pl.Table.from_panel_name_df_dict(
        {
            'One': EXAMPLE_DF,
            'Two': (EXAMPLE_DF + 10)
        },
        caption='My Table Title',
        below_text='My below text'
    )
    two_panel_table_from_lol_no_index = pl.Table.from_list_of_lists_of_dfs(
        [[EXAMPLE_DF], [EXAMPLE_DF + 10]],
        panel_names=['One', 'Two'],
        caption='My Table Title',
        below_text='My below text'
    )
    two_panel_table_from_dict_no_index_with_tl = pl.Table.from_panel_name_df_dict(
        {
            'One': EXAMPLE_DF,
            'Two': (EXAMPLE_DF + 10)
        },
        caption='My Table Title',
        below_text='My below text',
        top_left_corner_labels='woo'
    )
    two_panel_table_from_lol_no_index_with_tl = pl.Table.from_list_of_lists_of_dfs(
        [[EXAMPLE_DF], [EXAMPLE_DF + 10]],
        panel_names=['One', 'Two'],
        caption='My Table Title',
        below_text='My below text',
        top_left_corner_labels=['woo'],
    )
    two_panel_table_from_dict_with_index = pl.Table.from_panel_name_df_dict(
        {
            'One': EXAMPLE_DF,
            'Two': (EXAMPLE_DF + 10)
        },
        caption='My Table Title',
        below_text='My below text',
        include_index=True
    )
    two_panel_table_from_dict_with_index_and_name = pl.Table.from_panel_name_df_dict(
        {
            'One': EXAMPLE_DF_WITH_NAMED_INDEX,
            'Two': (EXAMPLE_DF_WITH_NAMED_INDEX + 10)
        },
        caption='My Table Title',
        below_text='My below text',
        include_index=True,
        align = 'lccc',
    )
    two_panel_table_from_dict_with_index_and_tl = pl.Table.from_panel_name_df_dict(
        {
            'One': EXAMPLE_DF,
            'Two': (EXAMPLE_DF + 10)
        },
        caption='My Table Title',
        below_text='My below text',
        include_index=True,
        top_left_corner_labels='TL Label'
    )
    two_panel_table_from_lol_with_index = pl.Table.from_list_of_lists_of_dfs(
        [[EXAMPLE_DF], [EXAMPLE_DF + 10]],
        panel_names=['One', 'Two'],
        caption='My Table Title',
        below_text='My below text',
        include_index=True
    )
    two_panel_table_from_lol_with_index_and_name = pl.Table.from_list_of_lists_of_dfs(
        [[EXAMPLE_DF_WITH_NAMED_INDEX], [EXAMPLE_DF_WITH_NAMED_INDEX + 10]],
        panel_names=['One', 'Two'],
        caption='My Table Title',
        below_text='My below text',
        include_index=True,
    )
    two_panel_table_from_lol_with_index_and_tl = pl.Table.from_list_of_lists_of_dfs(
        [[EXAMPLE_DF], [EXAMPLE_DF + 10]],
        panel_names=['One', 'Two'],
        caption='My Table Title',
        below_text='My below text',
        include_index=True,
        top_left_corner_labels='TL Label'
    )
    two_panel_table_from_dict_no_pad = pl.Table.from_panel_name_df_dict(
        {
            'One': EXAMPLE_DF,
            'Two': (EXAMPLE_DF + 10)
        },
        caption='My Table Title',
        below_text='My below text',
        include_index=True,
        pad_rows=0,
    )
    two_panel_table_from_dict_double_pad = pl.Table.from_panel_name_df_dict(
        {
            'One': EXAMPLE_DF,
            'Two': (EXAMPLE_DF + 10)
        },
        caption='My Table Title',
        below_text='My below text',
        include_index=True,
        pad_rows=2,
    )
    two_panel_table_from_lol_no_pad = pl.Table.from_list_of_lists_of_dfs(
        [[EXAMPLE_DF], [EXAMPLE_DF + 10]],
        panel_names=['One', 'Two'],
        caption='My Table Title',
        below_text='My below text',
        include_index=True,
        pad_rows=0,
        pad_columns=0,
    )
    two_panel_table_from_lol_double_pad = pl.Table.from_list_of_lists_of_dfs(
        [[EXAMPLE_DF], [EXAMPLE_DF + 10]],
        panel_names=['One', 'Two'],
        caption='My Table Title',
        below_text='My below text',
        include_index=True,
        pad_rows=2,
        pad_columns=0,
    )

    def test_table(self):
        assert str(self.table) == '\\begin{table}\n\\centering\n\\begin{threeparttable}\n\\caption{My Table Title}\n\\begin{tabular}{lcc}\n\\toprule\na & b & c\\\\\n 1 &  2 &  3 \\\\\n 4 &  5 &  6 \\\\\n 7 &  8 &  9 \\\\\n\\bottomrule\n\n\\end{tabular}\n\\begin{tablenotes}[para, flushleft]\nMy below text\n\\end{tablenotes}\n\\end{threeparttable}\n\\end{table}'

    def test_two_panel_table_no_index(self):
        assert str(self.two_panel_table_from_dict_no_index) == str(self.two_panel_table_from_lol_no_index) == '\\begin{table}\n\\centering\n\\begin{threeparttable}\n\\caption{My Table Title}\n\\begin{tabular}{lcc}\n\\toprule\na & b & c\\\\\n\\midrule\n\\multicolumn{3}{l}{Panel A: One}\\\\\n 1 &  2 &  3 \\\\\n 4 &  5 &  6 \\\\\n 7 &  8 &  9 \\\\\n  &   &  \\\\\n\\multicolumn{3}{l}{Panel B: Two}\\\\\n 11 &  12 &  13 \\\\\n 14 &  15 &  16 \\\\\n 17 &  18 &  19 \\\\\n\\bottomrule\n\n\\end{tabular}\n\\begin{tablenotes}[para, flushleft]\nMy below text\n\\end{tablenotes}\n\\end{threeparttable}\n\\end{table}'

    def test_two_panel_table_no_index_with_tl(self):
        assert str(self.two_panel_table_from_dict_no_index_with_tl) == str(self.two_panel_table_from_lol_no_index_with_tl) == '\\begin{table}\n\\centering\n\\begin{threeparttable}\n\\caption{My Table Title}\n\\begin{tabular}{lccc}\n\\toprule\nwoo & a & b & c\\\\\n\\midrule\n\\multicolumn{4}{l}{Panel A: One}\\\\\n &  1 &  2 &  3 \\\\\n &  4 &  5 &  6 \\\\\n &  7 &  8 &  9 \\\\\n  &   &   &  \\\\\n\\multicolumn{4}{l}{Panel B: Two}\\\\\n &  11 &  12 &  13 \\\\\n &  14 &  15 &  16 \\\\\n &  17 &  18 &  19 \\\\\n\\bottomrule\n\n\\end{tabular}\n\\begin{tablenotes}[para, flushleft]\nMy below text\n\\end{tablenotes}\n\\end{threeparttable}\n\\end{table}'

    def test_two_panel_table_with_index(self):
        assert str(self.two_panel_table_from_dict_with_index) == str(self.two_panel_table_from_lol_with_index) == '\\begin{table}\n\\centering\n\\begin{threeparttable}\n\\caption{My Table Title}\n\\begin{tabular}{lccc}\n\\toprule\n  & a & b & c\\\\\n\\midrule\n\\multicolumn{4}{l}{Panel A: One}\\\\\n0 &  1 &  2 &  3 \\\\\n1 &  4 &  5 &  6 \\\\\n2 &  7 &  8 &  9 \\\\\n  &   &   &  \\\\\n\\multicolumn{4}{l}{Panel B: Two}\\\\\n0 &  11 &  12 &  13 \\\\\n1 &  14 &  15 &  16 \\\\\n2 &  17 &  18 &  19 \\\\\n\\bottomrule\n\n\\end{tabular}\n\\begin{tablenotes}[para, flushleft]\nMy below text\n\\end{tablenotes}\n\\end{threeparttable}\n\\end{table}'

    def test_two_panel_table_with_index_and_name(self):
        assert str(self.two_panel_table_from_dict_with_index_and_name) == str(self.two_panel_table_from_lol_with_index_and_name) == '\\begin{table}\n\\centering\n\\begin{threeparttable}\n\\caption{My Table Title}\n\\begin{tabular}{lccc}\n\\toprule\n  & a & b & c\\\\\n\\midrule\n\\multicolumn{4}{l}{Panel A: One}\\\\\n0 &  1 &  2 &  3 \\\\\n1 &  4 &  5 &  6 \\\\\n2 &  7 &  8 &  9 \\\\\n  &   &   &  \\\\\n\\multicolumn{4}{l}{Panel B: Two}\\\\\n0 &  11 &  12 &  13 \\\\\n1 &  14 &  15 &  16 \\\\\n2 &  17 &  18 &  19 \\\\\n\\bottomrule\n\n\\end{tabular}\n\\begin{tablenotes}[para, flushleft]\nMy below text\n\\end{tablenotes}\n\\end{threeparttable}\n\\end{table}'

    def test_two_panel_table_with_index_and_tl(self):
        assert str(self.two_panel_table_from_dict_with_index_and_tl) == str(self.two_panel_table_from_lol_with_index_and_tl) == '\\begin{table}\n\\centering\n\\begin{threeparttable}\n\\caption{My Table Title}\n\\begin{tabular}{lccc}\n\\toprule\nTL Label & a & b & c\\\\\n\\midrule\n\\multicolumn{4}{l}{Panel A: One}\\\\\n0 &  1 &  2 &  3 \\\\\n1 &  4 &  5 &  6 \\\\\n2 &  7 &  8 &  9 \\\\\n  &   &   &  \\\\\n\\multicolumn{4}{l}{Panel B: Two}\\\\\n0 &  11 &  12 &  13 \\\\\n1 &  14 &  15 &  16 \\\\\n2 &  17 &  18 &  19 \\\\\n\\bottomrule\n\n\\end{tabular}\n\\begin{tablenotes}[para, flushleft]\nMy below text\n\\end{tablenotes}\n\\end{threeparttable}\n\\end{table}'

    def test_two_panel_table_no_pad(self):
        assert str(self.two_panel_table_from_dict_no_pad) == str(self.two_panel_table_from_lol_no_pad) == '\\begin{table}\n\\centering\n\\begin{threeparttable}\n\\caption{My Table Title}\n\\begin{tabular}{lccc}\n\\toprule\n  & a & b & c\\\\\n\\midrule\n\\multicolumn{4}{l}{Panel A: One}\\\\\n0 &  1 &  2 &  3 \\\\\n1 &  4 &  5 &  6 \\\\\n2 &  7 &  8 &  9 \\\\\n\\midrule\n\\multicolumn{4}{l}{Panel B: Two}\\\\\n0 &  11 &  12 &  13 \\\\\n1 &  14 &  15 &  16 \\\\\n2 &  17 &  18 &  19 \\\\\n\\bottomrule\n\n\\end{tabular}\n\\begin{tablenotes}[para, flushleft]\nMy below text\n\\end{tablenotes}\n\\end{threeparttable}\n\\end{table}'

    def test_two_panel_table_double_pad(self):
        assert str(self.two_panel_table_from_dict_double_pad) == str(self.two_panel_table_from_lol_double_pad) == '\\begin{table}\n\\centering\n\\begin{threeparttable}\n\\caption{My Table Title}\n\\begin{tabular}{lccc}\n\\toprule\n  & a & b & c\\\\\n\\midrule\n\\multicolumn{4}{l}{Panel A: One}\\\\\n0 &  1 &  2 &  3 \\\\\n1 &  4 &  5 &  6 \\\\\n2 &  7 &  8 &  9 \\\\\n  &   &   &  \\\\\n  &   &   &  \\\\\n\\multicolumn{4}{l}{Panel B: Two}\\\\\n0 &  11 &  12 &  13 \\\\\n1 &  14 &  15 &  16 \\\\\n2 &  17 &  18 &  19 \\\\\n\\bottomrule\n\n\\end{tabular}\n\\begin{tablenotes}[para, flushleft]\nMy below text\n\\end{tablenotes}\n\\end{threeparttable}\n\\end{table}'

    def test_table_to_tabular(self):
        tabular = self.two_panel_table_from_lol_with_index.tex_obj(as_document=False, as_single_tabular=True)
        assert str(tabular) == '\\begin{tabular}{lccc}\n\\toprule\n  & a & b & c\\\\\n\\midrule\n\\multicolumn{4}{l}{Panel A: One}\\\\\n0 &  1 &  2 &  3 \\\\\n1 &  4 &  5 &  6 \\\\\n2 &  7 &  8 &  9 \\\\\n  &   &   &  \\\\\n\\multicolumn{4}{l}{Panel B: Two}\\\\\n0 &  11 &  12 &  13 \\\\\n1 &  14 &  15 &  16 \\\\\n2 &  17 &  18 &  19 \\\\\n\\bottomrule\n\n\\end{tabular}'

    def test_table_to_tabular_list(self):
        tabular_list = self.two_panel_table_from_lol_with_index.tex_obj(as_document=False, as_panel_tabular_list=True)
        assert len(tabular_list) == 2
        assert isinstance(tabular_list, list)
        assert str(tabular_list[0]) == '\\begin{tabular}{lccc}\n\\toprule\n\\multicolumn{4}{l}{Panel A: One}\\\\\n  & a & b & c\\\\\n\\midrule\n0 &  1 &  2 &  3 \\\\\n1 &  4 &  5 &  6 \\\\\n2 &  7 &  8 &  9 \\\\\n\\bottomrule\n\n\\end{tabular}'
        assert str(tabular_list[1]) == '\\begin{tabular}{lccc}\n\\toprule\n\\multicolumn{4}{l}{Panel A: Two}\\\\\n  & a & b & c\\\\\n\\midrule\n0 &  11 &  12 &  13 \\\\\n1 &  14 &  15 &  16 \\\\\n2 &  17 &  18 &  19 \\\\\n\\bottomrule\n\n\\end{tabular}'

    def test_table_in_document(self):
        doc = pl.Document([self.table])
        assert str(doc) == '\\documentclass[]{article}\n\\usepackage{threeparttable}\n\\usepackage{booktabs}\n\\usepackage{dcolumn}\n\\newcolumntype{L}[1]{>{\\raggedright\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{C}[1]{>{\\centering\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{R}[1]{>{\\raggedleft\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{.}{D{.}{.}{-1}}\n\\usepackage{amsmath}\n\\usepackage{pdflscape}\n\\usepackage{array}\n\\usepackage{fancyhdr}\n\\usepackage{lastpage}\n\\usepackage{textcomp}\n\\usepackage[T1]{fontenc}\n\\usepackage{caption}\n\\usepackage{subcaption}\n\\usepackage{graphicx}\n\\usepackage[margin=0.8in, bottom=1.2in]{geometry}\n\\usepackage[page]{appendix}\n\\pagestyle{fancy}\n\\renewcommand{\\headrulewidth}{0pt}\n\\fancyhead{}\n\\rfoot{Page \\thepage\\  of \\pageref{LastPage}}\n\\cfoot{}\n\\begin{document}\n\\begin{table}\n\\centering\n\\begin{threeparttable}\n\\caption{My Table Title}\n\\begin{tabular}{lcc}\n\\toprule\na & b & c\\\\\n 1 &  2 &  3 \\\\\n 4 &  5 &  6 \\\\\n 7 &  8 &  9 \\\\\n\\bottomrule\n\n\\end{tabular}\n\\begin{tablenotes}[para, flushleft]\nMy below text\n\\end{tablenotes}\n\\end{threeparttable}\n\\end{table}\n\\end{document}'
        name = 'document with table'
        doc.to_pdf(outfolder=GENERATED_FILES_DIR, outname=name)
        compare_pdfs_in_generated_vs_input_by_name(name)

    def test_two_panel_table_in_document(self):
        doc = pl.Document([self.two_panel_table_from_lol_with_index])
        assert str(doc) == '\\documentclass[]{article}\n\\usepackage{threeparttable}\n\\usepackage{booktabs}\n\\usepackage{dcolumn}\n\\newcolumntype{L}[1]{>{\\raggedright\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{C}[1]{>{\\centering\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{R}[1]{>{\\raggedleft\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{.}{D{.}{.}{-1}}\n\\usepackage{amsmath}\n\\usepackage{pdflscape}\n\\usepackage{array}\n\\usepackage{fancyhdr}\n\\usepackage{lastpage}\n\\usepackage{textcomp}\n\\usepackage[T1]{fontenc}\n\\usepackage{caption}\n\\usepackage{subcaption}\n\\usepackage{graphicx}\n\\usepackage[margin=0.8in, bottom=1.2in]{geometry}\n\\usepackage[page]{appendix}\n\\pagestyle{fancy}\n\\renewcommand{\\headrulewidth}{0pt}\n\\fancyhead{}\n\\rfoot{Page \\thepage\\  of \\pageref{LastPage}}\n\\cfoot{}\n\\begin{document}\n\\begin{table}\n\\centering\n\\begin{threeparttable}\n\\caption{My Table Title}\n\\begin{tabular}{lccc}\n\\toprule\n  & a & b & c\\\\\n\\midrule\n\\multicolumn{4}{l}{Panel A: One}\\\\\n0 &  1 &  2 &  3 \\\\\n1 &  4 &  5 &  6 \\\\\n2 &  7 &  8 &  9 \\\\\n  &   &   &  \\\\\n\\multicolumn{4}{l}{Panel B: Two}\\\\\n0 &  11 &  12 &  13 \\\\\n1 &  14 &  15 &  16 \\\\\n2 &  17 &  18 &  19 \\\\\n\\bottomrule\n\n\\end{tabular}\n\\begin{tablenotes}[para, flushleft]\nMy below text\n\\end{tablenotes}\n\\end{threeparttable}\n\\end{table}\n\\end{document}'
        name = 'document with two panel table'
        doc.to_pdf(outfolder=GENERATED_FILES_DIR, outname=name)
        compare_pdfs_in_generated_vs_input_by_name(name)



from pyexlatex.models.item import MultiOptionSimpleItem, SimpleItem
from pyexlatex.models.sizes.textsizes import TextSize

class DeclareFloatFont(MultiOptionSimpleItem):
    name = 'DeclareFloatFont'

    def __init__(self, relative_size_int: int):
        self.size_def = TextSize(relative_size_int)
        super().__init__(self.name, self.size_def.name, self.size_def)


class FloatSetup(SimpleItem):
    name = 'floatsetup'

    def __init__(self, targeted_float: str, options: str):
        super().__init__(self.name, options, pre_modifiers=f'[{targeted_float}]')



from typing import Optional, Any, Sequence
from pyexlatex.models.item import MultiOptionSimpleItem
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.models.package import Package
from pyexlatex.exc import NoPackageWithNameException


class Hyperlink(TextAreaMixin, MultiOptionSimpleItem):
    """
    Create links to urls, with or without displaying text instead of the link.
    """

    def __init__(self, href: str, content: Optional[Any] = None, **kwargs):
        self.href = href
        self.content = content

        options: Sequence[Any]
        if content is None:
            self.name = 'url'
            options = (href,)
        else:
            self.name = 'href'
            options = (href, content)

        self.add_data_from_content(content)
        self.data.packages.append(
            # TODO [#14]: Figure out way to have different options for different hyperlinks
            #
            # Think about passing options. Difficult because mutliple Hyperlinks could be constructed with
            # different options, then not clear which to take.
            Package('hyperref', modifier_str='hidelinks')
        )

        MultiOptionSimpleItem.__init__(self, self.name, *options, **kwargs)



from typing import Union, List, Tuple, Dict, Optional, Sequence

from pyexlatex.models.item import Item, ItemBase, IsLatexItemMixin
from pyexlatex.models.documentitem import DocumentItem

AnyItem = ItemBase
ListOfItems = List[ItemBase]
DictOfItems = Dict[str, AnyItem]
ListOrDictOfItems = Union[ListOfItems, DictOfItems]
ListOrDictOrItem = Union[ListOrDictOfItems, AnyItem]
ItemOrListOfItems = Union[AnyItem, ListOfItems]
StrList = List[str]
StrListOrNone = Union[StrList, None]
BytesList = List[bytes]
BytesListOrNone = Optional[BytesList]
ItemAndPreEnvContents = Tuple[AnyItem, StrListOrNone]
PyexlatexItem = Union[IsLatexItemMixin, str]
PyexlatexItems = Union[PyexlatexItem, Sequence[PyexlatexItem]]


from typing import Optional, Sequence
from pyexlatex.models.item import ItemBase
from pyexlatex.models.references.bibtex.base import BibTexEntryBase
from pyexlatex.models.references.bibtex.style.manager import StyleManager
from pyexlatex.models.control.filecontents import FileContents
from pyexlatex.models.containeritem import ContainerItem
from pyexlatex.models.package import Package
from pyexlatex.texgen import _basic_item_str


class Bibliography(ContainerItem, ItemBase):
    """
    High-level class for working with bibliographies in LaTeX. Handles style as well as the references.
    Include an instance of this class where you want the bibliography printed.
    """

    def __init__(self, references: Optional[Sequence[BibTexEntryBase]] = None, style_name: str = 'plain'):
        self.references = references
        style_manager = StyleManager()
        self.style = style_manager.get_style_by_name(style_name)
        self.add_data_from_content(self.style)
        self._include_references_file(references)
        self.data.packages.append(Package('natbib'))

    def _include_references_file(self, references: Optional[Sequence[BibTexEntryBase]] = None):
        from pyexlatex.logic.builder import _build
        all_references = _build(references)
        references_inline_file = FileContents(all_references, 'refs.bib')
        self.data.end_document_items.append(references_inline_file)
        self.data.packages.append(Package('filecontents'))

    def __str__(self):
        from pyexlatex.logic.builder import _build
        bibliography_str = _basic_item_str('bibliography', 'refs')
        output = _build([
            self.style,
            bibliography_str
        ])
        return output

        



from typing import Union, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from pyexlatex.table import ColumnsAlignment


class BaseTabular:

    @staticmethod
    def _get_num_columns_from_content(content) -> int:
        # Content should be made a list before this function. Get first or only content back
        reference_content = content[0]

        # Try to get from num_columns attribute if it exists
        num_columns = getattr(reference_content, 'num_columns', None)
        if num_columns is not None:
            return num_columns

        # Otherwise, try to take the length of the first passed content
        return len(reference_content)

    def _get_columns_alignment_from_passed_align(self, content,
                                                 align: Optional[Union['ColumnsAlignment', str]] = None
                                                 ) -> 'ColumnsAlignment':
        from pyexlatex.table import ColumnsAlignment
        if align is None:
            return ColumnsAlignment(
                num_columns=BaseTabular._get_num_columns_from_content(content)
            )
        elif isinstance(align, str):
            return ColumnsAlignment.from_alignment_str(align)
        else:
            # Assumed to be passed ColumnsAlignment or something else that will directly work in its place
            return align


from pyexlatex.models.blindtext import BlindText
from tests.base import NoOptionsNoContentsItemTest


class TestBlindText(NoOptionsNoContentsItemTest):
    item_class = BlindText
    tag_name = 'blindtext'



import os
import subprocess
from subprocess import CalledProcessError
from typing import Optional, List, Sequence

from future.utils import raise_from
from data import Data as I
from data.decorators import data
from shutilwhich import which
from tempdir import TempDir

from latex.exc import LatexBuildError


class BaseBuilder:
    """
    The base class for LaTeX file builders
    """
    output_extension: str
    pre_file_output_args: Sequence[str] = (
        '-interaction=batchmode',
        '-halt-on-error',
        '-shell-escape',
        '-file-line-error'
    )
    post_file_output_args: Sequence[str] = tuple()
    default_executable: str

    def __init__(self, executable: Optional[str] = None, bibtex: str = 'bibtex', max_runs: int = 15):
        if executable is None:
            executable = self.default_executable

        self.executable = executable
        self.bibtex = bibtex
        self.max_runs = max_runs

    @data('source')
    def build(self, source, texinputs: Optional[List[str]] = None, run_bibtex: bool = False):
        if texinputs is None:
            texinputs = []

        with TempDir() as tmpdir,\
                source.temp_saved(suffix='.latex', dir=tmpdir) as tmp:
            # close temp file, so other processes can access it also on Windows
            tmp.close()
            called_bibtex = False

            # calculate output filename
            base_fn = os.path.splitext(tmp.name)[0]
            output_fn = base_fn + f'.{self.output_extension}'
            aux_fn = base_fn + '.aux'
            args = [self.executable, *self.pre_file_output_args, tmp.name, *self.post_file_output_args]

            # create environment
            newenv = os.environ.copy()
            inputs_value = os.pathsep.join(texinputs) + os.pathsep
            newenv['TEXINPUTS'] = inputs_value
            newenv['BSTINPUTS'] = inputs_value

            # run until aux file settles
            prev_aux = None
            runs_left = self.max_runs
            self._pre_compile(tmpdir, base_fn)
            while runs_left:
                try:
                    subprocess.check_call(args,
                                          cwd=tmpdir,
                                          env=newenv,
                                          stdin=open(os.devnull, 'r'),
                                          stdout=open(os.devnull, 'w'), )
                except CalledProcessError as e:
                    raise_from(LatexBuildError(base_fn + '.log'), e)

                # check aux-file
                aux = open(aux_fn, 'rb').read()

                if aux == prev_aux:
                    # Stable aux file
                    if run_bibtex and not called_bibtex:
                        called_bibtex = True  # ensure only called once
                        bibtex_args = [self.bibtex, os.path.basename(aux_fn)]
                        try:
                            subprocess.check_call(bibtex_args,
                                              cwd=tmpdir,
                                              env=newenv,
                                              stdin=open(os.devnull, 'r'),
                                              stdout=open(os.devnull, 'w'), )
                        except CalledProcessError as e:
                            # TODO [#9]: better handling for LaTeX exceptions
                            #
                            # Parse log file, raise proper exception
                            with open(base_fn + '.blg', 'r') as f:
                                log_contents = f.read()
                            raise Exception(log_contents)
                        continue  # go back into the loop to process with biblography
                    break

                prev_aux = aux
                runs_left -= 1
            else:
                raise RuntimeError(
                    'Maximum number of runs ({}) without a stable .aux file '
                    'reached.'.format(self.max_runs))

            return I(open(output_fn, 'rb').read(), encoding=None)

    def is_available(self):
        return bool(which(self.executable))

    def _pre_compile(self, temp_dir: str, base_file_name: str):
        pass



from pyexlatex.models.format.breaks import LineBreak

def build_figure_content(items, caption=None, label=None, centering=True, position_str=None):
    from pyexlatex.texgen import _centering_str
    elements = [
        position_str,
        _centering_str() if centering else None,
        *[str(item) for item in items],
        str(caption) if caption else None,
        str(label) if label else None
    ]

    valid_elements = [elem for elem in elements if elem]

    return _build(valid_elements)

def _build(elements):
    return LineBreak().join(elements)


def build(content):
    if hasattr(content, 'contents'):
        result = build(content.contents)
        content.contents = result
    if hasattr(content, 'content'):
        result = build(content.content)
        content.content = result
    if hasattr(content, 'contents') or hasattr(content, 'content'):
        return content
    elif isinstance(content, (list, tuple)):
        built = _build([build(c) for c in content])
        return built
    elif content is None:
        return ''
    else:
        return str(content)



from typing import List, Optional, Sequence, Union

from pyexlatex.exc import NoPackageWithNameException
from pyexlatex.models.documentitem import DocumentItem
from pyexlatex.models.package import Package
from mixins.repr import ReprMixin
from mixins.attrequals import EqOnAttrsMixin, EqHashMixin
from pyexlatex.models.references.bibtex.base import BibTexEntryBase


class UniqueDataList(list):
    """
    A list where the items are always unique.
    """

    def __init__(self, iterable):
        if iterable is None:
            iterable = []
        super().__init__(iterable)

    def append(self, value):
        if value is None:
            return
        if value in self:
            return
        return super().append(value)

    def extend(self, iterable):
        if iterable is None:
            return
        new_items = [item for item in iterable if item not in self]
        if not new_items:
            return
        return super().extend(new_items)


class UniquePackagesList(UniqueDataList):
    """
    A unique data list that allows passing of just a string to represent a package
    """

    def __init__(self, packages: Optional[List[Union[Package, str]]] = None):
        if packages is None:
            packages = []

        packages = self._as_packages(packages)

        super().__init__(packages)

    def _as_package(self, package: Union[Package, str]):
        if isinstance(package, str):
            return Package(package)

        return package

    def _as_packages(self, packages: Optional[List[Union[Package, str]]] = None):
        if packages is None:
            return packages
        return [self._as_package(pack) for pack in packages]

    def append(self, package: Union[Package, str]):
        if package is None:
            return
        package = self._as_package(package)
        super().append(package)

    def extend(self, packages: Optional[List[Union[Package, str]]] = None):  # type: ignore
        if packages is None:
            return
        packages = self._as_packages(packages)
        super().extend(packages)

    def get_by_name(self, name: str):
        for package in self:
            if package.matches_name(name):
                return package
        raise NoPackageWithNameException(name)

    def delete_by_name(self, name: str):
        new_packages = []
        for package in self:
            if not hasattr(package, 'matches_name'):
                # Got something other than a package
                # TODO [#13]: think about how to delete non-packages from packages
                new_packages.append(package)  # No way to check if name matches, so just keep it
            elif not package.matches_name(name):
                new_packages.append(package)
        if len(new_packages) == len(self):
            raise NoPackageWithNameException(name)

        super().__init__(new_packages)


class DocumentSetupData(ReprMixin, EqOnAttrsMixin, EqHashMixin):
    repr_cols = [
        'filepaths',
        'begin_document_items',
        'end_document_items',
        'packages',
        'source_paths',
        'references',
        'flags'
    ]
    ignore_attrs = [
        'ignore_attrs',
        'repr_cols',
        'attrs',
        'readable_repr',
        'equal_attrs'
    ]
    equal_attrs = [
        'filepaths',
        'binaries',
        'begin_document_items',
        'end_document_items',
        'packages',
        'source_paths',
        'references',
        'flags'
    ]
    non_unique_attrs = [
        'filepaths',
        'binaries'
    ]

    def __init__(self, filepaths: Optional[List[str]] = None, binaries: Optional[List[bytes]] = None, 
                 begin_document_items: Optional[List[DocumentItem]] = None, 
                 end_document_items: Optional[List[DocumentItem]] = None, packages: Optional[List[Union[Package, str]]] = None,
                 source_paths: Optional[List[str]] = None, references: Optional[Sequence[BibTexEntryBase]] = None,
                 flags: Optional[List[str]] = None):
        self.filepaths = UniqueDataList(filepaths)
        self.binaries = UniqueDataList(binaries)
        self.begin_document_items = UniqueDataList(begin_document_items)
        self.end_document_items = UniqueDataList(end_document_items)
        self.packages = UniquePackagesList(packages)
        self.source_paths = UniqueDataList(source_paths)
        self.references = UniqueDataList(references)
        self.flags = UniqueDataList(flags)

    @property
    def attrs(self) -> List[str]:
        return [attr for attr in dir(self) if (not attr.startswith('_') and not attr in self.ignore_attrs)]


from typing import Optional, Union, List
from pyexlatex.models.template import Template
import pyexlatex as pl
from pyexlatex.models.format.paragraph.multicol import MultiColumn


class SpacedSection(Template):
    """
    A section which appropriately spaces content for a resume.
    """

    def __init__(self, contents, title: str, end_adjustment: float = -0.2,
                 num_cols: int = 1):
        self.end_adjustment = end_adjustment
        self.num_cols = num_cols

        if not isinstance(contents, (list, tuple)):
            contents = [contents]

        all_contents: Union[List[Union[pl.VSpace, str]], MultiColumn] = [pl.VSpace(0.2)]
        for content in contents:
            all_contents.append(content)
            all_contents.append('')
        all_contents[-1] = pl.VSpace(end_adjustment)  # replace final spacing

        if self.num_cols > 1:
            all_contents = MultiColumn(all_contents, num_cols=self.num_cols)

        self.contents = pl.Section(all_contents, title=title)
        super().__init__()



from pyexlatex import LetterDocument
from pyexlatex.letter.closing import Closing
from pyexlatex.letter.enclosures import Enclosures
from pyexlatex.letter.letter import Letter
from pyexlatex.letter.opening import Opening
from pyexlatex.letter.ps import PS
from pyexlatex.letter.signature import Signature
import pyexlatex as pl
from tests.base import TextAreaTest, EnvironmentTest


class TestClosing(TextAreaTest):
    area_class = Closing
    tag_name = 'closing'


class TestEnclosures(TextAreaTest):
    area_class = Enclosures
    tag_name = 'encl'


class TestOpening(TextAreaTest):
    area_class = Opening
    tag_name = 'opening'


class TestPS(TextAreaTest):
    area_class = PS
    tag_name = 'ps'


class TestSignature(TextAreaTest):
    area_class = Signature
    tag_name = 'signature'


class TestLetter(EnvironmentTest):
    env_class = Letter
    tag_name = 'letter'


class TestLetterDocument:
    no_options_result = '\\documentclass[]{letter}\n\\longindentation=0pt\n\\begin{document}\n\\begin{letter}{}\n\\opening{Dear Sir or Madam:}\nwoo\n\\closing{Sincerely,}\n\\end{letter}\n\\end{document}'

    def test_no_options_str(self):
        doc = LetterDocument('woo')
        assert str(doc) == self.no_options_result

    def test_no_options_list(self):
        doc = LetterDocument(['woo'])
        assert str(doc) == self.no_options_result

    def test_auto_package_inclusion(self):
        doc = LetterDocument(['woo', pl.RGB(10, 20, 30, color_name='cool')])
        assert str(doc) == '\\documentclass[]{letter}\n\\usepackage{xcolor}\n\\longindentation=0pt\n\\begin{document}\n\\begin{letter}{}\n\\opening{Dear Sir or Madam:}\nwoo\n\\definecolor{cool}{RGB}{10,20,30}\n\\closing{Sincerely,}\n\\end{letter}\n\\end{document}'

    def test_all_options(self):
        doc = LetterDocument(
            'woo',
            contact_info=['my name','123 main st', 'my city, state'],
            to_contact_info=['your name', '456 main st', 'their city, state'],
            signer_name='my name twice',
            closing_indent='5pt',
            salutation='yo girl:',
            closing='yours forever and ever',
            ps='an extra thought!',
            enclosures=['one thing', 'another thing'],
        )
        assert str(doc) == '\\documentclass[]{letter}\n\\address{my name \\\\ 123 main st \\\\ my city, state}\n\\signature{my name twice}\n\\longindentation=5pt\n\\begin{document}\n\\begin{letter}{your name \\\\ 456 main st \\\\ their city, state}\n\\opening{yo girl:}\nwoo\n\\closing{yours forever and ever}\n\\ps{an extra thought!}\n\\encl{one thing\nanother thing}\n\\end{letter}\n\\end{document}'



from typing import List

import pandas as pd

from pyexlatex.logic.tools import _max_len_or_zero
from pyexlatex.table.models.panels.grid import PanelGrid, GridShape
from mixins.repr import ReprMixin
from pyexlatex.table.models.data.table import DataTable
from pyexlatex.table.models.table.row import Row
from pyexlatex.table.models.table.section import TableSection



class Panel(ReprMixin):
    """
    Represents one section of a table. May have multiple DataTables within one panel.


    """
    repr_cols = ['name', 'panel_grid']

    def __init__(self, panel_grid: PanelGrid, name: str=None):
        self.panel_grid = panel_grid
        self.name = name

    @classmethod
    def from_data_tables(cls, data_table_list: List[DataTable], shape: tuple=None, name: str=None):
        """

        :param data_table_list: list of pyexlatex.table.DataTable
        :param shape: tuple of (rows, columns) to arrange DataTables. They will be placed from left to right,
                      then from top to bottom.
                      passsing None defaults one column, as many rows as DataTables
        :param name: name to be displayed with panel
        :return:
        """
        return cls(PanelGrid(data_table_list, shape=shape), name=name)

    @classmethod
    def from_df(cls, df: pd.DataFrame, include_columns=True, include_index=False, name: str=None):
        data_table = DataTable.from_df(
            df,
            include_columns=include_columns,
            include_index=include_index
        )

        return cls.from_data_tables([data_table], name=name)

    @classmethod
    def from_df_list(cls, df_list: List[pd.DataFrame], shape: tuple=None, name: str=None, include_columns=True,
                     include_index=False, data_table_kwargs={}):
        """

        :param df_list: list of pandas DataFrame
        :param shape: tuple of (rows, columns) to arrange DataFrames. They will be placed from left to right,
                      then from top to bottom.
                      passsing None defaults one column, as many rows as DataFrames
        :param name: name to be displayed with panel
        :param include_columns:
        :param include_index:
        :param data_table_kwargs: kwargs to be passed to DataTable.from_df. Same kwargs will be passed to
                                  all data tables.
        :return:
        """
        data_table_list = [
            DataTable.from_df(
                df,
                include_columns=include_columns,
                include_index=include_index,
                **data_table_kwargs
            )
            for df in df_list
        ]

        return cls.from_data_tables(data_table_list, shape=shape, name=name)

    @property
    def rows(self):
        try:
            return self._rows
        except AttributeError:
            self._rows = self._set_rows()

        return self._rows

    def _set_rows(self):
        from pyexlatex.table.models.labels.table import LabelTable, LabelCollection
        from pyexlatex.table.models.labels.label import Label
        rows: [Row] = []

        # Add panel name
        if self.name is not None:
            label_collection = LabelCollection([Label(self.name, span=self.num_columns, align='l')])
            name_table = LabelTable([label_collection])
            rows += name_table.rows

        for grid_row in self.panel_grid:
            for i, table_section in enumerate(grid_row):
                if i == 0:
                    panel_row = table_section
                else:
                    panel_row += table_section
            rows += panel_row.rows
        return rows

    @property
    def is_spacer(self):
        return all([row.is_spacer for row in self.rows])

    @property
    def num_columns(self):
        num_columns = 0
        for grid_row in self.panel_grid:
            row_num_columns = [table_section.num_columns for table_section in grid_row]
            max_for_row = max(row_num_columns)
            num_columns = max(num_columns, max_for_row) # update global max with row max
        return num_columns









from pyexlatex.models.section.base import TextAreaBase


class Columns(TextAreaBase):
    name = 'columns'
    repr_cols = ['contents']

    def __init__(self, content, **kwargs):
        super().__init__(self.name, content, **kwargs)



import os
from typing import Type
import pathlib

from pyexlatex import Italics, EnvironmentTemplate, Footnote
from pyexlatex.letter.letter import Letter
from pyexlatex.models.environment import Environment
from pyexlatex.models.format.centering import Centering
from pyexlatex.models.item import MultiOptionSimpleItem
from pyexlatex.models.landscape import Landscape

TESTS_DIR = pathlib.Path(__file__).parent
INPUT_FILES_DIR = TESTS_DIR / 'input_files'
EXAMPLE_IMAGE_PATH = INPUT_FILES_DIR / 'nd-logo.png'
GENERATED_FILES_DIR = TESTS_DIR / 'generated_files'

if not os.path.exists(GENERATED_FILES_DIR):
    os.makedirs(GENERATED_FILES_DIR)


class NoOptionsNoContentsItemTest:
    item_class: Type = Centering
    tag_name: str = 'centering'

    def test_item(self):
        content = self.item_class()
        assert str(content) == '\\' + self.tag_name


class SimpleItemTest:
    item_class: Type = Footnote
    tag_name: str = 'footnote'

    def test_item(self):
        content = self.item_class('woo')
        assert str(content) == '\\' + self.tag_name + '{woo}'


class MultiOptionSimpleItemTest:
    item_class: Type = MultiOptionSimpleItem
    tag_name: str = 'footnote'

    def test_one_item(self):
        content = self.item_class('woo')
        assert str(content) == '\\' + self.tag_name + '{woo}'

    def test_two_items(self):
        content = self.item_class('woo', 'yeah')
        assert str(content) == '\\' + self.tag_name + '{woo}'


class TextAreaTest:
    area_class: Type = Italics
    tag_name: str = 'section'

    def test_str(self):
        content = self.area_class('woo')
        assert str(content) == '\\' + self.tag_name + '{woo}'

    def test_list(self):
        content = self.area_class(['woo', 'yeah'])
        assert str(content) == '\\' + self.tag_name + '{woo\nyeah}'


class EnvironmentTest:
    env_class: Type = Letter
    tag_name: str = 'environment'

    def test_no_modifiers(self):
        content = self.env_class().wrap('woo')
        braces_tag_name = f'{{{self.tag_name}}}'
        assert str(content) == f'\\begin{braces_tag_name}\nwoo\n\end{braces_tag_name}'

    def test_modifiers(self):
        modifier_str = '{stuff}'
        content = self.env_class(modifiers=modifier_str).wrap('woo')
        braces_tag_name = f'{{{self.tag_name}}}'
        assert str(content) == f'\\begin{braces_tag_name}{modifier_str}\nwoo\n\end{braces_tag_name}'


from typing import List, Optional, Callable, cast

from pyexlatex.models.format.breaks import OutputLineBreak
from pyexlatex.models.format.raw import Raw
from pyexlatex.letter.closing import Closing
from pyexlatex.letter.enclosures import Enclosures
from pyexlatex.letter.letter import Letter
from pyexlatex.letter.opening import Opening
from pyexlatex.letter.ps import PS
from pyexlatex.letter.signature import Signature
from pyexlatex.models.document import DocumentBase
from pyexlatex.resume.models.contactline import ContactLine
from pyexlatex.typing import ItemOrListOfItems, PyexlatexItems, PyexlatexItem
from pyexlatex.models.package import Package
from pyexlatex.models.control.documentclass.documentclass import DocumentClass


class LetterDocument(DocumentBase):
    """
    The main high-level document class for creating letters.
    """
    name = 'document'

    def __init__(self, content: PyexlatexItems, contact_info: Optional[PyexlatexItems] = None,
                 to_contact_info: PyexlatexItems = '',
                 signer_name: Optional[PyexlatexItems] = None, closing_indent: str = '0pt',
                 salutation: PyexlatexItems = 'Dear Sir or Madam:', closing: Optional[PyexlatexItems] = 'Sincerely,',
                 ps: Optional[PyexlatexItems] = None, enclosures: Optional[PyexlatexItems] = None,
                 packages: List[Package] = None,
                 pre_env_contents: Optional[ItemOrListOfItems] = None,
                 font_size: Optional[float] = None, doc_class_options: Optional[List[str]] = None,
                 pre_output_func: Optional[Callable] = None):

        self.document_class_obj = DocumentClass(
            document_type='letter',
            font_size=font_size,
            options=doc_class_options
        )

        self.add_data_from_content(
            [content, contact_info, to_contact_info, signer_name, salutation, closing, ps, enclosures]
        )

        all_pre_env_contents: List[PyexlatexItems] = []

        if contact_info is not None:
            all_pre_env_contents.append(ContactLine(contact_info))
        if signer_name is not None:
            all_pre_env_contents.append(Signature(signer_name))
        all_pre_env_contents.append(Raw(fr'\longindentation={closing_indent}'))

        if pre_env_contents is not None:
            all_pre_env_contents.extend(pre_env_contents)

        letter_contents = [
            Opening(salutation),
            content,
        ]

        if closing is not None:
            letter_contents.append(Closing(closing))
        if ps is not None:
            letter_contents.append(PS(ps))
        if enclosures is not None:
            letter_contents.append(Enclosures(enclosures))

        if isinstance(to_contact_info, (list, tuple)):
            to_contact_info = f' {OutputLineBreak()} '.join(to_contact_info)
        to_contact_info = cast(PyexlatexItem, to_contact_info)

        letter = Letter(modifiers=self._wrap_with_braces(to_contact_info)).wrap(letter_contents)

        super().__init__(
            letter,
            packages=packages,
            pre_env_contents=all_pre_env_contents,  # type: ignore
            pre_output_func=pre_output_func,
        )





from typing import Optional, Sequence, Union, List
import pyexlatex as pl
from pyexlatex.models.template import Template


class Reference(Template):
    SPACE_BETWEEN_ADJUSTMENT = '-8pt'

    def __init__(self, name: str, title_lines: Optional[Sequence[str]] = None, company: Optional[str] = None,
                 contact_lines: Optional[Sequence[str]] = None, email: Optional[str] = None,
                 prevent_page_break: bool = True):
        self.name = name
        self.title_lines = title_lines
        self.company = company
        self.contact_lines = contact_lines
        self.email = email
        self.prevent_page_break = prevent_page_break
        self.contents = self._get_contents()
        super().__init__()

    def _get_contents(self):
        if self.contact_lines:
            contact_lines = self.contact_lines
        else:
            contact_lines = []
        if self.title_lines:
            title_lines = [pl.Italics(line) for line in self.title_lines]
        else:
            title_lines = []

        possible_contents = [
            pl.SmallCaps(pl.Bold(self.name)),
            *title_lines,
            self.company,
            *contact_lines,
            self.email,
            pl.OutputLineBreak(size_adjustment=self.SPACE_BETWEEN_ADJUSTMENT),
        ]

        out_contents = []
        for content in possible_contents:
            out_contents.append(content)
            out_contents.append(pl.OutputLineBreak())  # force line breaks in content
        del out_contents[-1]  # remove last line break

        if self.prevent_page_break:
            out_contents = pl.NoPageBreak(out_contents)

        return out_contents



class AmpersandAddMixin:

    def __add__(self, other):
        if hasattr(self, 'value') and hasattr(other, 'value'):
            return self.value + ' & ' + other.value
        elif hasattr(self, 'values') and hasattr(other, 'value'):
            return ' & '.join(self.values + [other.value])
        elif hasattr(self, 'values') and hasattr(other, 'values'):
            return ' & '.join(self.values + other.values)
        elif hasattr(self, 'value') and hasattr(other, 'values'):
            return ' & '.join([self.value] + other.values)
        elif hasattr(self, 'value'):
            return self.value + ' & ' + other
        elif hasattr(self, 'values'):
            return ' & '.join(self.values + [other])
        return self + ' & ' + other

    def __radd__(self, other):
        if hasattr(self, 'value') and hasattr(other, 'value'):
            return other.value + ' & ' + self.value
        elif hasattr(self, 'values') and hasattr(other, 'value'):
            return ' & '.join([other.value] + self.values)
        elif hasattr(self, 'values') and hasattr(other, 'values'):
            return ' & '.join(other.values + self.values)
        elif hasattr(self, 'value') and hasattr(other, 'values'):
            return ' & '.join(other.values + [self.value])
        elif hasattr(self, 'value'):
            return other + self.value + ' & '
        elif hasattr(self, 'values'):
            return ' & '.join([other] + self.values)
        return other + ' & ' + self


from mixins import EqOnAttrsMixin

from pyexlatex.models.item import IsLatexItemMixin, IsSpecificClassMixin


class Raw(IsSpecificClassMixin, IsLatexItemMixin, EqOnAttrsMixin):
    r"""
    Don't replace latex characters in this blocks such as \ or %, useful for adding latex commands or comments manually.
    """
    equal_attrs = ('contents',)

    def __init__(self, contents):
        self.contents = contents
        super().__init__()

    def __str__(self):
        return self.contents



from pyexlatex.models.item import SimpleItem, MultiOptionSimpleItem
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.typing import PyexlatexItems


class PageStyle(SimpleItem):
    """
    Sets style of the pages for the entire document.
    """
    name = 'pagestyle'

    def __init__(self, contents: PyexlatexItems):
        super().__init__(self.name, contents)


class ThisPageStyle(SimpleItem):
    """
    Sets style of the current page in a document.
    """
    name = 'thispagestyle'

    def __init__(self, contents: PyexlatexItems):
        super().__init__(self.name, contents)


class FancyPageStyle(TextAreaMixin, MultiOptionSimpleItem):
    """
    Redefines an existing page style such as plain but with fancy page style features.
    """
    name = 'fancypagestyle'

    def __init__(self, contents: PyexlatexItems, page_style: str = 'plain'):
        TextAreaMixin.__init__(self, self.name, contents)
        str_contents = self._build(self.contents)
        MultiOptionSimpleItem.__init__(self, self.name, page_style, str_contents)
